# -*- coding: utf-8 -*-
"""
This module provides a class :class:`Expr` to subclass from in order to
describe expressions. The hope was that this class would allow straightforward
interoperability between python packages handling symbolics (SymPy) and units
(quantities) as well as working without either of those. The price for this has
been a complicated implementation and unnatural syntax.

Note that this module is to be considered an implementation detail, and not
something that should be relied upon in external code. Consider the ``.util._expr``
module deprecated and aim to use SymPy expressions where possible, and handle units
separately.
"""
from __future__ import (absolute_import, division, print_function)

import math
from itertools import chain
from operator import add, mul, truediv, sub, pow
from ..units import is_quantity, unit_of, to_unitless
from .pyutil import defaultkeydict, deprecated


try:
    import sympy
except ImportError:
    sympy = None


def _implicit_conversion(obj):
    if isinstance(obj, (int, float)) or is_quantity(obj):
        return Constant(obj)
    elif isinstance(obj, Expr):
        return obj
    elif isinstance(obj, str):
        return Symbol(unique_keys=(obj,))

    if sympy is not None:
        if isinstance(obj, sympy.Mul):
            if len(obj.args) != 2:
                raise NotImplementedError("Did you use evaluate=False?")
            return _MulExpr([_implicit_conversion(obj.args[0]), _implicit_conversion(obj.args[1])])
        elif isinstance(obj, sympy.Add):
            if len(obj.args) != 2:
                raise NotImplementedError("Did you use evaluate=False?")
            return _AddExpr([_implicit_conversion(obj.args[0]), _implicit_conversion(obj.args[1])])
        elif isinstance(obj, sympy.Pow):
            return _PowExpr(_implicit_conversion(obj.base), _implicit_conversion(obj.exp))
        elif isinstance(obj, sympy.Float):
            return Constant(float(obj))
        elif isinstance(obj, sympy.Symbol):
            return Symbol(unique_keys=(obj.name,))
        elif isinstance(obj, sympy.Number):
            return Constant(float(obj))

    raise NotImplementedError(
        "Don't know how to convert %s (of type %s)" % (obj, type(obj)))


class Expr(object):
    ''' Baseclass for Expressions corresponding to physical quantities.

    The design assumes that a large group of different Expr subclasses may
    be evaluated with some shared state (parameter_keys). The backend kwarg
    in call enables use of e.g. math, numpy or sympy interchangeably.

    Parameters
    ----------
    args : tuple/list of scalars or dict mapping name to scalar
        When dict: it is converted to a list using ``self.argument_names`` or
        ``self.unique_keys``.
    unique_keys : iterable of strings
        Unique names (among all instances) for late overriding, aligned with beginning of
        ``args``.

    Examples
    --------
    >>> class HeatCapacity(Expr):
    ...     parameter_keys = ('temperature',)
    ...
    >>> import math
    >>> class EinsteinSolid(HeatCapacity):
    ...     parameter_keys = HeatCapacity.parameter_keys + ('molar_gas_constant',)
    ...     argument_names = ('einstein_temperature', 'molar_mass')
    ...
    ...     def __call__(self, variables, backend=math):
    ...         TE, molar_mass = self.all_args(variables, backend=backend)  # einstein_temperature
    ...         T, R = self.all_params(variables, backend=backend)
    ...         # Canonical ensemble:
    ...         molar_c_v = 3*R*(TE/(2*T))**2 * backend.sinh(TE/(2*T))**-2
    ...         return molar_c_v/molar_mass
    ...
    >>> from chempy import Substance
    >>> Al = Substance.from_formula('Al', data={'DebyeT': 428})
    >>> Be = Substance.from_formula('Be', data={'DebyeT': 1440})
    >>> einT = lambda s: 0.806*s.data['DebyeT']
    >>> cv = {s.name: EinsteinSolid([einT(s), s.mass]) for s in (Al, Be)}
    >>> print('%.4f' % cv['Al']({'temperature': 273.15, 'molar_gas_constant': 8.3145}))  # J/(g*K)
    0.8108
    >>> import sympy; from sympy import Symbol as Symb
    >>> print(cv['Be']({'temperature': Symb('T'), 'molar_gas_constant': Symb('R')}, backend=sympy))
    112105.346283965*R/(T**2*sinh(580.32/T)**2)

    Attributes
    ----------
    argument_names : tuple of strings, optional
        For documentation and referencing positional arguments in self.args
        If set, and `nargs` is `None`: its length is used to set `nargs`
        (unless argument_names ends with an Ellipsis()).
    argument_defaults : tuple of floats, optional
        Default values for arguments, aligned from the end of argument names.
    parameter_keys : tuple of strings
    nargs : int
        number of arguments (`None` signifies unset, -1 signifies any number)
    '''

    argument_names = None
    argument_defaults = None
    parameter_keys = ()
    nargs = None

    @property
    def trivially_zero(self):
        return False

    def __init__(self, args=None, unique_keys=None):
        if isinstance(args, str):
            args = (args,)
        if self.argument_names is not None and self.argument_names[-1] != Ellipsis and self.nargs is None:
            self.nargs = len(self.argument_names)
        if self.argument_defaults is not None:
            if self.nargs == -1:
                raise ValueError("Cannot have defaults when number of arguments is unbounded.")
            if len(self.argument_defaults) > len(self.argument_names):
                raise ValueError("Cannot have more defaults than actual arguments")
            if args is not None:
                n_missing = self.nargs - len(args)
                if n_missing > 0:
                    args = tuple(chain(args, self.argument_defaults[-n_missing:]))

        if self.nargs == 1 and (isinstance(args, (float, int)) or
                                getattr(args, 'ndim', -1) == 0 or
                                isinstance(args, Expr)):
            args = [args]
            nargs = 1
        elif args is None:
            nargs = None
        else:
            nargs = len(args)

        if self.nargs not in (None, -1) and nargs is not None and nargs != self.nargs:
            raise ValueError("Incorrect number of arguments: %d (expected %d)" % (nargs, self.nargs))
        if unique_keys is not None and self.nargs is not None and len(unique_keys) > self.nargs:
            raise ValueError("Incorrect number of unique_keys: %d (expected %d or less)" % (
                len(unique_keys), self.nargs))
        self.unique_keys = None if unique_keys is None else tuple(unique_keys)

        if isinstance(args, dict):
            args = [args[k] for k in self.argument_names or self.unique_keys]

        self.args = args

    @classmethod
    def fk(cls, *args):
        """ Alternative constructor "from keys", \\*args is used as ``unique_keys``. """
        return cls(unique_keys=args)

    @classmethod
    def from_callback(cls, callback, attr='__call__', **kwargs):
        """ Factory of subclasses

        Parameters
        ----------
        callback : callable
            signature: *args, backend=math
        attr : str
            What attribute to override
        argument_names : tuple of str, optional
        argument_defaults : tuple of floats, optional
        parameter_keys : tuple of str, optional,
        nargs : int, optional

        Examples
        --------
        >>> from operator import add; from functools import reduce
        >>> def poly(args, x, backend=math):
        ...     x0 = args[0]
        ...     return reduce(add, [c*(x-x0)**i for i, c in enumerate(args[1:])])
        ...
        >>> Poly = Expr.from_callback(poly, parameter_keys=('x',), argument_names=('x0', Ellipsis))
        >>> p = Poly([1, 3, 2, 5])
        >>> p({'x': 7}) == 3 + 2*(7-1) + 5*(7-1)**2
        True
        >>> q = Poly([1, 3, 2, 5], unique_keys=('x0_q',))
        >>> q({'x': 7, 'x0_q': 0}) == 3 + 2*7 + 5*7**2
        True

        """
        def body(self, variables, backend=math, **kw):
            args = self.all_args(variables, backend=backend)
            params = self.all_params(variables, backend=backend)
            return callback(args, *params, backend=backend, **kw)

        class Wrapper(cls):
            pass

        setattr(Wrapper, attr, body)
        Wrapper.__name__ = callback.__name__
        for k, v in kwargs.items():
            setattr(Wrapper, k, v)
        return Wrapper

    def __call__(self, variables, backend=math, **kwargs):
        raise NotImplementedError("Subclass and implement __call__")

    def __float__(self):
        try:
            return float(self({}))
        except KeyError:
            raise ValueError("Missing values for variables")

    def _all_keys(self, attr):
        _keys = getattr(self, attr)
        _all = set() if _keys is None else set(_keys)
        if self.args is not None:
            for arg in self.args:
                if isinstance(arg, Expr):
                    _all = _all.union(arg._all_keys(attr))
        return _all

    def all_parameter_keys(self):
        return self._all_keys('parameter_keys')

    def all_unique_keys(self):
        return self._all_keys('unique_keys')

    def _str(self, arg_fmt, unique_keys_fmt=str):
        if self.args is None or len(self.args) == 0:
            args_str = ''
        elif len(self.args) == 1:
            args_str = '%s,' % self.args[0]
        else:
            args_str = '%s' % ', '.join(map(arg_fmt, self.args))
        args_strs = [', '.join(chain(
            ['(%s)' % args_str],
            [unique_keys_fmt(self.unique_keys)] if self.unique_keys is not None else []
        ))]
        return "{}({})".format(self.__class__.__name__, ', '.join(args_strs))

    def __repr__(self):
        return self._str(repr)

    def string(self, arg_fmt=str, **kwargs):
        return self._str(arg_fmt, **kwargs)

    def arg(self, variables, index, backend=math, evaluate=True, **kwargs):
        """
        Parameters
        ----------
        variables : container
        index : int or str
            When str: index from ``self.argument_names``.
        backend : module
        evaluate : bool

        Notes
        -----
        Priority:
            1. unique_keys
            2. variables[k] for k in argument_names
        """
        if isinstance(index, str):
            index = self.argument_names.index(index)

        if self.unique_keys is None:
            res = self.args[index]
        elif index < len(self.unique_keys):
            uk = self.unique_keys[index]
            try:
                res = variables[uk]
            except KeyError:
                if self.args is None:
                    raise KeyError("Unique key missing: %s" % uk)
                else:
                    res = self.args[index]
        else:
            if self.args is None or index > len(self.args):
                res = self.argument_defaults[index - self.nargs + len(self.argument_defaults)]
            else:
                res = self.args[index]

        if isinstance(res, str):
            res = variables[res]
        # elif isinstance(res, Symbol):
        #     res = variables[res.unique_keys[0]]

        if isinstance(res, Expr) and evaluate:
            return res(variables, backend=backend, **kwargs)
        else:
            return res

    def all_args(self, variables, backend=math, evaluate=True, **kwargs):
        if self.nargs is None or self.nargs == -1:
            nargs = len(self.args)
        else:
            nargs = self.nargs
        return [self.arg(variables, i, backend, evaluate, **kwargs) for i in range(nargs)]

    def all_params(self, variables, backend=math):
        return [v(variables, backend=backend) if isinstance(v, Expr) else v for v
                in [variables[k] for k in self.parameter_keys]]

    def args_dimensionality(self, **kwargs):
        """ return tuple of dicts mapping str to int ('length', 'mass', 'time', 'current',
        'temperature', 'luminous_intensity', 'amount') """
        raise NotImplementedError("method not implemented in subclass.")

    def dedimensionalisation(self, unit_registry, variables={}, backend=math):
        """ Create an instance with consistent units from a unit_registry

        Parameters
        ----------
        unit_registry : dict
        variables : dict
        backend : module

        Examples
        --------
        >>> class Pressure(Expr):
        ...     argument_names = ('n',)
        ...     parameter_keys = ('temperature', 'volume', 'R')
        ...     def __call__(self, variables, backend=math, **kwargs):
        ...         n, = self.all_args(variables, backend=backend)
        ...         T, V, R = self.all_params(variables, backend=backend)
        ...         return n*R*T/V
        ...
        >>> from chempy.units import SI_base_registry, default_units as u
        >>> p = Pressure([2*u.micromole])
        >>> units, d = p.dedimensionalisation(SI_base_registry)
        >>> bool(units[0] == 1e6*u.micromole)
        True
        >>> assert d.args[0] == 2e-6


        Returns
        -------
        new_units: list of units of the dedimensionalised args.
        self.__class__ instance: with dedimensioanlised arguments

        """
        from ..units import default_unit_in_registry, to_unitless, unitless_in_registry
        new_units = []
        if self.args is None:
            unitless_args = None
        else:
            unitless_args = []
            units = [None if isinstance(arg, Expr) else default_unit_in_registry(arg, unit_registry) for arg
                     in self.all_args(variables, backend=backend, evaluate=False)]
            for arg, unit in zip(self.all_args(variables, backend=backend, evaluate=False), units):
                if isinstance(arg, Expr):
                    if unit is not None:
                        raise ValueError()
                    _unit, _dedim = arg.dedimensionalisation(unit_registry, variables, backend=backend)
                else:
                    _unit, _dedim = unit, to_unitless(arg, unit)
                new_units.append(_unit)
                unitless_args.append(_dedim)
        instance = self.__class__(unitless_args, self.unique_keys)
        if self.argument_defaults is not None:
            instance.argument_defaults = tuple(unitless_in_registry(arg, unit_registry)
                                               for arg in self.argument_defaults)

        return new_units, instance

    def _sympy_format(self, method, variables, backend, default, **kwargs):
        variables = variables or {}
        if backend in (None, math):
            backend = sympy
        variables = defaultkeydict(
            None if default is None else (lambda k: backend.Symbol(default(k))),
            {k: v if isinstance(v, Expr) else (backend.Symbol(v) if isinstance(v, str) else backend.Float(v))
             for k, v in variables.items()})
        expr = self(variables, backend=backend, **kwargs).simplify()
        if method == 'latex':
            return backend.latex(expr)
        elif method == 'str':
            return str(expr)
        elif method == 'unicode':
            return backend.pretty(expr, use_unicode=True)
        elif method == 'mathml':
            from sympy.printing.mathml import mathml
            return mathml(expr)
        else:
            raise NotImplementedError("Unknown method: %s" % method)

    def latex(self, variables=None, backend=math, default=None):
        r"""
        Parameters
        ----------
        variables : dict
        backend : module
        default : callable
            Format string based on missing key, signature: str -> str.

        Examples
        --------
        >>> def pressure(args, *params, **kw):
        ...     return args[0]*params[0]*params[1]/params[2]
        >>> Pressure = Expr.from_callback(pressure, parameter_keys='R temp vol'.split(), nargs=1)
        >>> p = Pressure([7])
        >>> p.latex({'R': 'R', 'temp': 'T', 'vol': 'V'})  # doctest: +SKIP
        '\\frac{7 R T}{V}'

        Notes
        -----
        Requires SymPy

        """
        return self._sympy_format('latex', variables, backend=backend, default=default)

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        if self.args is None and other.args is None:
            for uk1, uk2 in zip(self.unique_keys, other.unique_keys):
                if uk1 != uk2:
                    return False
            return True
        if self.args is None or other.args is None:
            return False
        if len(self.args) != len(other.args):
            return False
        from ..units import compare_equality
        for arg0, arg1 in zip(self.args, other.args):
            if not compare_equality(arg0, arg1):
                return False
        return True

    def __add__(self, other):
        _other = _implicit_conversion(other)
        if _other.trivially_zero:
            return self
        return _AddExpr.make(self, _other)

    def __sub__(self, other):
        if other == other*0:
            return self
        return _SubExpr.make(self, _implicit_conversion(other))

    def __mul__(self, other):
        if other == 1:
            return self
        if isinstance(other, UnaryWrapper):
            return NotImplemented  # delegate to UnaryWrapper.__rmul__
        return _MulExpr.make(self, _implicit_conversion(other))

    def __truediv__(self, other):
        if other == 1:
            return self
        if isinstance(other, UnaryWrapper):
            return NotImplemented  # delegate to UnaryWrapper.__rtruediv__
        return _DivExpr.make(self, _implicit_conversion(other))

    def __neg__(self):
        if isinstance(self, _NegExpr):
            return self.args[0]
        return _NegExpr.make(self)

    def __abs__(self):
        if isinstance(self, _AbsExpr):
            return self
        return _AbsExpr.make(self)

    def __radd__(self, other):
        return self+other

    def __rmul__(self, other):
        return self*other

    def __rsub__(self, other):
        return (-self) + other

    def __rtruediv__(self, other):
        return _DivExpr.make(_implicit_conversion(other), self)

    def __pow__(self, other):
        return _PowExpr.make(self, _implicit_conversion(other))

    def __rpow__(self, other):
        return _PowExpr.make(_implicit_conversion(other), self)


class UnaryWrapper(Expr):

    def __checks(self):
        if self.nargs != 1:
            raise ValueError("UnaryWrapper can only be used when nargs == 1")
        if self.unique_keys is not None:
            raise ValueError("UnaryWrapper can only be used when unique_keys are None")

    def __mul__(self, other):
        self.__checks()
        arg, = self.args
        return self.__class__([_MulExpr([arg, _implicit_conversion(other)])])

    def __truediv__(self, other):
        if other == 1:
            return self
        self.__checks()
        arg, = self.args
        return self.__class__([_DivExpr([arg, _implicit_conversion(other)])])

    def __rtruediv__(self, other):
        self.__checks()
        arg, = self.args
        return self.__class__([_DivExpr([_implicit_conversion(other), arg])])

    @classmethod
    def from_callback(cls, callback, attr='__call__', **kwargs):
        Wrapper = super().from_callback(callback, attr=attr, **kwargs)
        return lambda *args, **kw: cls(Wrapper(*args, **kw))


class _UnaryExpr(Expr):
    _op = None

    @classmethod
    def make(cls, arg, *args, **kwargs):
        assert not isinstance(arg, (tuple, list))
        if isinstance(arg, Constant) and isinstance(arg.args[0], (float, int)):
            inner_arg = arg.args[0]
            new_args = cls._op(inner_arg),
            return Constant(new_args)
        else:
            return cls((arg,), *args, **kwargs)


class _NegExpr(_UnaryExpr):

    def _op(x):
        return -x

    def _str(self, *args, **kwargs):
        return "-%s" % args[0]._str(*args, **kwargs)

    def __repr__(self):
        return super(_NegExpr, self)._str(repr)

    def __call__(self, variables, backend=math, **kwargs):
        arg0, = self.all_args(variables, backend=backend, **kwargs)
        return -arg0

    def rate_coeff(self, *args, **kwargs):   # <--- feature creep into base-class...
        return -self.args[0].rate_coeff(*args, **kwargs),


class _AbsExpr(_UnaryExpr):
    _op = abs

    def _str(self, *args, **kwargs):
        return "abs(%s)" % args[0]._str(*args, **kwargs)

    def __repr__(self):
        return super(_AbsExpr, self)._str(repr)

    def __call__(self, variables, backend=math, **kwargs):
        arg0, = self.all_args(variables, backend=backend, **kwargs)
        return abs(arg0)


class _BinaryExpr(Expr):
    _op = None

    @classmethod
    def make(cls, a1, a2, *args, **kwargs):
        assert not isinstance(a1, (tuple, list))
        assert not isinstance(a2, (tuple, list))
        if all(isinstance(arg, Constant) and isinstance(arg.args[0], (float, int)) for arg in [a1, a2]):
            return Constant(cls._op(a1.args[0], a2.args[0]), *args, **kwargs)
        else:
            return cls((a1, a2), *args, **kwargs)

    def _str(self, *args, **kwargs):
        return ("({0} %s {1})" % self._op_str).format(*[arg._str(*args, **kwargs) for arg in self.args])

    def __repr__(self):
        return super(_BinaryExpr, self)._str(repr)

    def __call__(self, variables, backend=math, **kwargs):
        arg0, arg1 = self.all_args(variables, backend=backend, **kwargs)
        return self._op(arg0, arg1)

    def rate_coeff(self, *args, **kwargs):
        return self._op(self.args[0].rate_coeff(*args, **kwargs),
                        self.args[1].rate_coeff(*args, **kwargs))


class _AddExpr(_BinaryExpr):
    _op = add
    _op_str = '+'


class _SubExpr(_BinaryExpr):
    _op = sub
    _op_str = '-'


class _MulExpr(_BinaryExpr):
    _op = mul
    _op_str = '*'

    @property
    def trivially_zero(self):
        try:
            return self.args[0].trivially_zero or self.args[1].trivially_zero
        except Exception:
            return False


class _DivExpr(_BinaryExpr):
    _op = truediv
    _op_str = '/'


class _PowExpr(_BinaryExpr):
    _op = pow
    _op_str = '**'


class Constant(Expr):
    nargs = 1

    @property
    def trivially_zero(self):
        return self.args[0] == 0

    def __call__(self, variables, backend=None, **kwargs):
        return self.args[0]

    def rate_coeff(self, *args, **kwargs):
        return self.args[0]

    def __lt__(self, other):
        if isinstance(other, (float, int)) and isinstance(self.args[0], (float, int)):
            return self.args[0] < other
        elif isinstance(other, Constant):
            return self < other.args[0]
        else:
            return NotImplemented


class Symbol(Expr):
    nargs = 1

    def _str(self, *args, **kwargs):
        uk, = self.unique_keys
        return uk

    def __repr__(self):
        return super(Symbol, self)._str(repr)

    def __call__(self, variables, backend=None, **kwargs):
        uk, = self.unique_keys
        return variables[uk]


class Function(Expr):
    pass


class UnaryFunction(Function):
    nargs = 1
    _func_name = None

    def __call__(self, variables, backend=math, **kwargs):
        arg, = self.all_args(variables, backend=backend, **kwargs)
        return getattr(backend, self._func_name)(arg)

    def rate_coeff(self, *args, **kwargs):
        return getattr(kwargs.get('backend', math), self._func_name)(self.args[0].rate_coeff(*args, **kwargs))


class BinaryFunction(Function):
    nargs = 2
    _func_name = None


class Log10(UnaryFunction):
    _func_name = 'log10'


class Exp(UnaryFunction):
    _func_name = 'exp'


def create_Piecewise(parameter_name, nan_fallback=False):
    """
    Examples
    --------
    >>> Power = Expr.from_callback(lambda args, x, backend=None: args[0]*x**args[1],
    ...     argument_names=('scale', 'pow'), parameter_keys=('x',))
    >>> minus_x = Power([-1, 1])
    >>> cube = Power([1, 3])
    >>> PW = create_Piecewise('x')
    >>> pw = PW([-float('inf'), minus_x, 0, cube, float('inf')])
    >>> pw({'x': -5}) == 5
    True
    >>> pw({'x': 2}) == 8
    True

    """
    def _pw(bounds_exprs, x, backend=math, **kwargs):
        if len(bounds_exprs) < 3:
            raise ValueError("Need at least 3 args")
        if len(bounds_exprs) % 2 != 1:
            raise ValueError("Need an odd number of bounds/exprs")
        n_exprs = (len(bounds_exprs) - 1) // 2
        lower = [bounds_exprs[2*(i+0)] for i in range(n_exprs)]
        upper = [bounds_exprs[2*(i+1)] for i in range(n_exprs)]
        exprs = [bounds_exprs[2*i + 1] for i in range(n_exprs)]

        if is_quantity(x):
            xunit = unit_of(x)

            def ulx(arg):
                return to_unitless(arg, xunit)
            lower = list(map(ulx, lower))
            upper = list(map(ulx, upper))
        else:
            # xunit = 1
            def ulx(arg):
                return arg

        try:
            pw = backend.Piecewise
        except AttributeError:
            for lo, up, ex in zip(lower, upper, exprs):
                if lo <= x <= up:
                    return ex
            else:
                raise ValueError("not within any bounds: %s" % x)
        else:
            _NAN = backend.Symbol('NAN')
            ux = ulx(x)
            return pw(
                *([(ex, backend.And(backend.Le(lo, ux), backend.Le(ux, up))) for
                   lo, up, ex in zip(lower, upper, exprs)] +
                  ([(_NAN, True)] if nan_fallback else []))
                , evaluate=False
            )

    return Expr.from_callback(_pw, parameter_keys=(parameter_name,))


def create_Poly(parameter_name, reciprocal=False, shift=None, name=None):
    """
    Examples
    --------
    >>> Poly = create_Poly('x')
    >>> p1 = Poly([3, 4, 5])
    >>> p1({'x': 7}) == 3 + 4*7 + 5*49
    True
    >>> RPoly = create_Poly('T', reciprocal=True)
    >>> p2 = RPoly([64, 32, 16, 8])
    >>> p2({'T': 2}) == 64 + 16 + 4 + 1
    True
    >>> SPoly = create_Poly('z', shift=True)
    >>> p3 = SPoly([7, 2, 3, 5], unique_keys=('z0',))
    >>> p3({'z': 9}) == 2 + 3*(9-7) + 5*(9-7)**2
    True
    >>> p3({'z': 9, 'z0': 6}) == 2 + 3*(9-6) + 5*(9-6)**2
    True

    """
    if shift is True:
        shift = 'shift'

    def _poly(args, x, backend=math, **kwargs):

        if shift is None:
            coeffs = args
            x0 = x
        else:
            coeffs = args[1:]
            x_shift = args[0]
            x0 = x - x_shift

        cur = 1
        res = None
        for coeff in coeffs:
            if res is None:
                res = coeff*cur
            else:
                res = res + coeff*cur

            if reciprocal:
                cur = cur / x0
            else:
                cur = cur * x0
        return res

    if shift is None:
        argument_names = None
    else:
        argument_names = (shift, Ellipsis)
    if name is not None:
        _poly.__name__ = name
    return Expr.from_callback(_poly, parameter_keys=(parameter_name,), argument_names=argument_names)

from ._expr_deprecated import _mk_PiecewisePoly, _mk_Poly  # noqa

mk_PiecewisePoly = deprecated(use_instead=create_Piecewise)(_mk_PiecewisePoly)
mk_Poly = deprecated(use_instead=create_Poly)(_mk_Poly)



"""
Interactive kinetics app with sliders.
Start by running:
    $ bokeh serve interactive.py
Add --show argument or navigate to:
    http://localhost:5006/interactive
"""
from collections import defaultdict
import sys

from chempy import Reaction, ReactionSystem
from chempy.arrhenius import ArrheniusParam
from chempy.util.bkh import integration_with_sliders


def get_rsys(Af=1e16, Ab=1.5e15, Ea=72e3, Er=-12e3):
    fw = Reaction({"Fe+3", "SCN-"}, {"FeSCN+2"}, ArrheniusParam(Af, Ea))
    bw = Reaction({"FeSCN+2"}, {"Fe+3", "SCN-"}, ArrheniusParam(Ab, Ea - Er))
    return ReactionSystem([fw, bw], "Fe+3 SCN- FeSCN+2".split())


if __name__.startswith("bk_"):
    from bokeh.io import curdoc

    curdoc().add_root(
        integration_with_sliders(
            get_rsys(),
            tend=3,
            c0=defaultdict(float, {"Fe+3": 3e-3, "SCN-": 1.5e-3, "FeSCN+2": 0.1e-3}),
            parameters={"temperature": 298.15},
            slider_kwargs={"temperature": dict(start=273.15, end=313.15, step=0.05)},
        )
    )
elif __name__ == "__main__":
    import warnings

    warnings.warn("Run using 'bokeh serve %s'" % __file__)
    sys.exit(1)



# -*- coding: utf-8 -*-
"""
This module collects object representing rate expressions. It is based
on the ``chemp.util._expr`` module. The API is somewhat cumbersome since
it tries to be compatible with pure python, SymPy and the underlying
units library of ChemPy (``quantities``). Consider the API to be provisional.
"""


from collections import OrderedDict
from functools import reduce
import math
from operator import add

from ..units import get_derived_unit, default_units, energy, concentration
from ..util._dimensionality import dimension_codes, base_registry
from ..util.pyutil import memoize, deprecated
from ..util._expr import Expr, UnaryWrapper, Symbol


_molar = getattr(default_units, "molar", 1)  # makes module importable.


class RateExpr(Expr):
    """Baseclass for rate expressions, see source code of e.g. MassAction & Radiolytic."""

    @classmethod
    @deprecated(use_instead=Expr.from_callback)
    def subclass_from_callback(cls, cb, cls_attrs=None):
        """Override RateExpr.__call__

        Parameters
        ----------
        cb : callback
            With signature (variables, all_args, backend) -> scalar
            where `variables` is a dict, `all_args` a tuple and `backend` a module.
        cls_attrs : dict, optional
            Attributes to set in subclass, e.g. parameter_keys, nargs

        Examples
        --------
        >>> from chempy import Reaction
        >>> rxn = Reaction({'O2': 1, 'H2': 1}, {'H2O2': 1})  # d[H2O2]/dt = p0*exp(-p1/T)*sqrt([O2])
        >>> def cb(variables, all_args, backend):
        ...     O2, T = variables['O2'], variables['temperature']
        ...     p0, p1 = all_args
        ...     return p0*backend.sqrt(O2)*backend.exp(-p1/T)
        >>> MyRateExpr = RateExpr.subclass_from_callback(cb, dict(parameter_keys=('temperature',),nargs=2))
        >>> k = MyRateExpr([1.3e9, 4317.2])
        >>> print('%.5g' % k({'temperature': 298.15, 'O2': 1.1e-3, 'rxn': rxn}))
        22.186

        """

        class _RateExpr(cls):
            def __call__(self, variables, backend=math, **kwargs):
                return cb(
                    variables, self.all_args(variables), backend=backend, **kwargs
                )

        for k, v in (cls_attrs or {}).items():
            setattr(_RateExpr, k, v)
        return _RateExpr


class RadiolyticBase(RateExpr):
    pass  # for isinstance checks


@memoize(None)
def mk_Radiolytic(*doserate_names):
    """Create a Radiolytic rate expression

    Note that there is no mass-action dependence in the resulting
    class, i.e. the rates does not depend on any concentrations.

    Parameters
    ----------
    \\*doserate_names : str instances
        Default: ('',)


    Examples
    --------
    >>> RadiolyticAlpha = mk_Radiolytic('alpha')
    >>> RadiolyticGamma = mk_Radiolytic('gamma')
    >>> dihydrogen_alpha = RadiolyticAlpha([0.8e-7])
    >>> dihydrogen_gamma = RadiolyticGamma([0.45e-7])
    >>> RadiolyticAB = mk_Radiolytic('alpha', 'beta')

    Notes
    -----
    The instance __call__ will require by default ``'density'`` and ``'doserate'``
    in variables.

    """
    if len(doserate_names) == 0:
        doserate_names = ("",)

    class _Radiolytic(RadiolyticBase):
        argument_names = tuple(
            "radiolytic_yield{0}".format("" if drn == "" else "_" + drn)
            for drn in doserate_names
        )  # [amount/energy]
        parameter_keys = ("density",) + tuple(
            "doserate{0}".format("" if drn == "" else "_" + drn)
            for drn in doserate_names
        )

        def args_dimensionality(self, reaction):
            N = base_registry["amount"]
            E = get_derived_unit(base_registry, "energy")
            return (dict(zip(dimension_codes, N / E)),) * self.nargs

        def g_values(self, *args, **kwargs):
            return OrderedDict(
                zip(self.parameter_keys[1:], self.all_args(*args, **kwargs))
            )

        @deprecated(use_instead="Radiolytic.all_args")
        def g_value(self, variables, backend=math, **kwargs):
            (g_val,) = self.all_args(variables, backend=backend, **kwargs)
            return g_val

        def __call__(self, variables, backend=math, reaction=None, **kwargs):
            return variables["density"] * reduce(
                add,
                [
                    variables[k] * gval
                    for k, gval in zip(
                        self.parameter_keys[1:],
                        self.all_args(variables, backend=backend, **kwargs),
                    )
                ],
            )

    _Radiolytic.__name__ = (
        "Radiolytic"
        if doserate_names == ("",)
        else ("Radiolytic_" + "_".join(doserate_names))
    )
    return _Radiolytic


Radiolytic = mk_Radiolytic()


class MassAction(RateExpr, UnaryWrapper):
    """Rate-expression of mass-action type

    Notes
    -----
    :meth:`__call__` requires a :class:`Reaction` instance to be passed as ``reaction``
    keyword argument.

    Examples
    --------
    >>> ma = MassAction([3.14])
    >>> from chempy import Reaction
    >>> r = Reaction.from_string('3 A -> B', param=ma)
    >>> r.rate({'A': 2}) == {'A': -75.36, 'B': 25.12}
    True

    """

    def _str(self, *args, **kwargs):
        (arg,) = self.args
        if isinstance(arg, Symbol):
            (uk,) = arg.unique_keys
            return "'%s'" % uk
        else:
            return super(MassAction, self)._str(*args, **kwargs)

    def __repr__(self):
        return super(MassAction, self)._str(repr)

    def get_named_keys(self):
        # Symbol uses args[0] to return from variables
        (arg,) = self.args
        if isinstance(arg, Symbol):
            return arg.args
        else:
            return self.unique_keys

    argument_names = ("rate_constant",)

    def args_dimensionality(self, reaction):
        order = reaction.order()
        return ({"time": -1, "amount": 1 - order, "length": 3 * (order - 1)},)

    def active_conc_prod(self, variables, backend=math, reaction=None):
        result = 1
        for k, v in reaction.reac.items():
            result = result * variables[k] ** v
        return result

    def rate_coeff(self, variables, backend=math, **kwargs):
        (rat_c,) = self.all_args(variables, backend=backend, **kwargs)
        return rat_c

    def __call__(self, variables, backend=math, reaction=None, **kwargs):
        return self.rate_coeff(
            variables, backend=backend, reaction=reaction
        ) * self.active_conc_prod(
            variables, backend=backend, reaction=reaction, **kwargs
        )

    def string(self, *args, **kwargs):
        if self.args is None and len(self.unique_keys) == 1:
            return self.unique_keys[0]
        else:
            return super(MassAction, self).string(*args, **kwargs)

    @classmethod
    @deprecated(use_instead="MassAction.from_callback")
    def subclass_from_callback(cls, cb, cls_attrs=None):
        """Override MassAction.__call__"""
        _RateExpr = super(MassAction, cls).subclass_from_callback(
            cb, cls_attrs=cls_attrs
        )

        def wrapper(*args, **kwargs):
            obj = _RateExpr(*args, **kwargs)
            return cls(obj)

        return wrapper

    @classmethod
    def from_callback(cls, callback, attr="__call__", **kwargs):
        Wrapper = RateExpr.from_callback(callback, attr=attr, **kwargs)
        return lambda *args, **kwargs: MassAction(Wrapper(*args, **kwargs))


class Arrhenius(Expr):
    """Rate expression for a Arrhenius-type of rate: c0*exp(-c1/T)

    Examples
    --------
    >>> from math import exp
    >>> from chempy import Reaction
    >>> from chempy.units import allclose, default_units as u
    >>> A = 1e11 / u.second
    >>> Ea_over_R = 42e3/8.3145 * u.K**-1
    >>> ratex = MassAction(Arrhenius([A, Ea_over_R]))
    >>> rxn = Reaction({'R'}, {'P'}, ratex)
    >>> dRdt = rxn.rate({'R': 3*u.M, 'temperature': 298.15*u.K})['R']
    >>> assert allclose(dRdt, -3*1e11*exp(-42e3/8.3145/298.15)*u.M/u.s)
    """

    argument_names = ("A", "Ea_over_R")
    parameter_keys = ("temperature",)

    def args_dimensionality(self, reaction):
        order = reaction.order()
        return (
            {"time": -1, "amount": 1 - order, "length": 3 * (order - 1)},
            {"temperature": 1},
        )

    def __call__(self, variables, backend=math, **kwargs):
        A, Ea_over_R = self.all_args(variables, backend=backend, **kwargs)
        try:
            Ea_over_R = Ea_over_R.simplified
        except AttributeError:
            pass
        return A * backend.exp(-Ea_over_R / variables["temperature"])


class Eyring(Expr):
    """Rate expression for Eyring eq: c0*T*exp(-c1/T)

    Note that choice of standard state (c^0) will matter for order > 1.
    """

    argument_names = ("kB_h_times_exp_dS_R", "dH_over_R", "conc0")
    argument_defaults = (1 * _molar,)
    parameter_keys = ("temperature",)

    def args_dimensionality(self, reaction):
        order = reaction.order()
        return (
            {
                "time": -1,
                "temperature": -1,
                "amount": 1 - order,
                "length": 3 * (order - 1),
            },
            {"temperature": 1},
            concentration,
        )

    def __call__(self, variables, backend=math, **kwargs):
        c0, c1, conc0 = self.all_args(variables, backend=backend, **kwargs)
        T = variables["temperature"]
        return c0 * T * backend.exp(-c1 / T) * conc0 ** (1 - kwargs["reaction"].order())


class EyringHS(Expr):
    argument_names = ("dH", "dS", "c0")
    argument_defaults = (1 * _molar,)
    parameter_keys = (
        "temperature",
        "molar_gas_constant",
        "Boltzmann_constant",
        "Planck_constant",
    )

    def args_dimensionality(self, **kwargs):
        return (
            energy + {"amount": -1},
            energy + {"amount": -1, "temperature": -1},
            concentration,
        )

    def __call__(self, variables, backend=math, reaction=None, **kwargs):
        dH, dS, c0 = self.all_args(variables, backend=backend, **kwargs)
        T, R, kB, h = [variables[k] for k in self.parameter_keys]
        return (
            kB
            / h
            * T
            * backend.exp(-(dH - T * dS) / (R * T))
            * c0 ** (1 - reaction.order())
        )


class RampedTemp(Expr):
    """Ramped temperature, pass as substitution to e.g. ``get_odesys``"""

    argument_names = ("T0", "dTdt")
    parameter_keys = ("time",)  # consider e.g. a parameter such as 'init_time'

    def args_dimensionality(self, **kwargs):
        return ({"temperature": 1}, {"temperature": 1, "time": -1})

    def __call__(self, variables, backend=None, **kwargs):
        T0, dTdt = self.all_args(variables, backend=backend, **kwargs)
        return T0 + dTdt * variables["time"]


class SinTemp(Expr):
    argument_names = ("Tbase", "Tamp", "angvel", "phase")
    parameter_keys = ("time",)

    def args_dimensionality(self, **kwargs):
        return ({"temperature": 1}, {"temperature": 1}, {"time": -1}, {})

    def __call__(self, variables, backend=math, **kwargs):
        Tbase, Tamp, angvel, phase = self.all_args(variables, backend=backend, **kwargs)
        return Tbase + Tamp * backend.sin(angvel * variables["time"] + phase)



# -*- coding: utf-8 -*-

from collections import OrderedDict
from .pyutil import defaultnamedtuple

dimension_codes = OrderedDict(
    zip(
        "length mass time current temperature amount".split(),  # not considering luminous_intensity
        "L M T I Θ N".split(),
    )
)


class DimensionalitySI(
    defaultnamedtuple(
        "DimensionalitySIBase", dimension_codes.keys(), (0,) * len(dimension_codes)
    )
):
    def __mul__(self, other):
        return self.__class__(*(x + y for x, y in zip(self, other)))

    def __truediv__(self, other):
        return self.__class__(*(x - y for x, y in zip(self, other)))

    def __pow__(self, exp):
        return self.__class__(*(x * exp for x in self))


base_registry = {name: DimensionalitySI(**{name: 1}) for name in dimension_codes}



# -*- coding: utf-8 -*-


from .._util import prodpow
from ..util.testing import requires


def test_prodpow():
    result = prodpow([11, 13], [[0, 1], [1, 2]])
    assert result[0] == 13
    assert result[1] == 11 * 13 * 13


@requires("sympy")
def test_prodpow__symbols():
    import sympy

    a, b = sympy.symbols("a b")
    exprs = prodpow([a, b], [[0, 1], [1, 2]])
    assert exprs[0] == b
    assert exprs[1] == a * b ** 2



from collections import OrderedDict
from chempy import Reaction
from chempy.kinetics.rates import MassAction, RadiolyticBase
from chempy.units import to_unitless, default_units as u


def jl_dict(od):
    return "Dict([%s])" % ", ".join(["(:%s, %.4g)" % (k, v) for k, v in od.items()])


def _r(r, p, substmap, parmap, *, unit_conc, unit_time, variables=None):
    """
    Parameters
    ==========
    ...
    variables: dict
        e.g. dict(doserate=99.9*u.Gy/u.s, density=998*u.kg/u.m3)
    """
    (pk,) = r.param.unique_keys
    if isinstance(r.param, MassAction):
        ratcoeff = to_unitless(p[pk], unit_conc ** (1 - r.order()) / unit_time)
        if not r.inact_reac:
            r_str = "{}, {}".format(
                parmap[pk],
                r.string(
                    substances=substmap,
                    with_param=False,
                    Reaction_arrow="-->",
                    Reaction_coeff_space="",
                ),
            )
        else:
            all_keys = r.keys()
            reac_stoichs = r.all_reac_stoich(all_keys)
            act_stoichs = r.active_reac_stoich(all_keys)
            rate = "*".join(
                [parmap[pk]]
                + [
                    ("%s^%d" % (substmap[k], v)) if v > 1 else substmap[k]
                    for k, v in zip(all_keys, act_stoichs)
                    if v > 0
                ]
            )
            r2 = Reaction(
                dict([(k, v) for k, v in zip(all_keys, reac_stoichs) if v]), r.prod
            )
            r_str = "{}, {}".format(
                rate,
                r2.string(
                    substances=substmap,
                    with_param=False,
                    Reaction_arrow="\u21D2",
                    Reaction_coeff_space="",
                ),
            )
    elif isinstance(r.param, RadiolyticBase):
        ratcoeff = to_unitless(
            p[pk] * variables["doserate"] * variables["density"], unit_conc / unit_time
        )
        assert not r.reac and not r.inact_reac and not r.inact_prod
        ((prod, n),) = r.prod.items()
        assert n == 1
        r_str = ("{}, 0 \u21D2 {}" if ratcoeff > 0 else "{}, {} \u21D2 0").format(
            parmap[pk], substmap[prod]
        )
    else:
        raise NotImplementedError("What's that?")
    return r_str, pk, abs(ratcoeff)


class DiffEqBioJl:
    _template_body = """\
{name} = @{crn_macro} begin
    {reactions}
end {parameters}
{post}
"""

    defaults = dict(unit_conc=u.molar, unit_time=u.second)

    def __init__(self, *, rxs, pars, substance_key_map, parmap, **kwargs):
        self.rxs = rxs
        self.pars = pars
        self.substance_key_map = substance_key_map
        self.parmap = parmap
        self.unit_conc = kwargs.get("unit_conc", self.defaults["unit_conc"])
        self.unit_time = kwargs.get("unit_time", self.defaults["unit_time"])

    @classmethod
    def from_rsystem(
        cls,
        rsys,
        par_vals,
        *,
        variables=None,
        substance_key_map=lambda i, sk: "y%d" % i,
        **kwargs
    ):
        if not isinstance(substance_key_map, dict):
            substance_key_map = {
                sk: substance_key_map(si, sk) for si, sk in enumerate(rsys.substances)
            }
        parmap = dict(
            [(r.param.unique_keys[0], "p%d" % i) for i, r in enumerate(rsys.rxns)]
        )
        rxs, pars = [], OrderedDict()
        for r in rsys.rxns:
            rs, pk, pv = _r(
                r,
                par_vals,
                substance_key_map,
                parmap,
                variables=variables,
                unit_conc=kwargs.get("unit_conc", cls.defaults["unit_conc"]),
                unit_time=kwargs.get("unit_time", cls.defaults["unit_time"]),
            )
            rxs.append(rs)
            if pk in pars:
                raise ValueError("Are you sure (sometimes intentional)?")
            pars[parmap[pk]] = pv
        return cls(
            rxs=rxs,
            pars=pars,
            substance_key_map=substance_key_map,
            parmap=parmap,
            **kwargs
        )

    def render_body(self, sparse_jac=False):
        name = "rn"
        return self._template_body.format(
            crn_macro="min_reaction_network" if sparse_jac else "reaction_network",
            name=name,
            reactions="\n    ".join(self.rxs),
            parameters=" ".join(self.pars),
            post="addodes!({}, sparse_jac=True)".format(name) if sparse_jac else "",
        )

    def render_setup(self, *, ics, atol, tex=True, tspan=None):
        export = ""
        export += "p = %s\n" % jl_dict(self.pars)
        export += "ics = %s\n" % jl_dict(
            OrderedDict(
                {
                    self.substance_key_map[k]: v
                    for k, v in to_unitless(ics, u.molar).items()
                }
            )
        )
        if atol:
            export += "abstol_d = %s\n" % jl_dict(
                {
                    self.substance_key_map[k]: v
                    for k, v in to_unitless(atol, u.molar).items()
                }
            )
            export += (
                "abstol = Array([get(abstol_d, k, 1e-10) for k=keys(speciesmap(rn))])"
            )
        if tex:
            export += "subst_tex = Dict([%s])\n" % ", ".join(
                '(:%s, ("%s", "%s"))' % (v, k, k.latex_name)
                for k, v in self.substance_key_map.items()
            )
        if tspan:
            export += """\
tspan = (0., %12.5g)
u0 = Array([get(ics, k, 1e-28) for k=keys(speciesmap(rn))])
parr = Array([p[k] for k=keys(paramsmap(rn))])
oprob = ODEProblem(rn, u0, tspan, parr)
"""
        return export

    def render_solve(self):
        return (
            "sol = solve(oprob, reltol=1e-9, abstol=abstol, Rodas5(),"
            " callback=PositiveDomain(ones(length(u0)), abstol=abstol))"
        )



# -*- coding: utf-8 -*-
"""
Non-public API (classes in this module may change without notice).

The purpose here is to define conventions, e.g. lower-case string
 'temperature' is used, opposed to e.g. 'T', 'Temperature', etc.
"""

from ..util._expr import create_Poly, create_Piecewise

TPoly = create_Poly("temperature")
RTPoly = create_Poly("temperature", reciprocal=True)
Log10TPoly = create_Poly("log10_temperature")
ShiftedTPoly = create_Poly("temperature", shift="Tref", name="ShiftedTPoly")
ShiftedLog10TPoly = create_Poly("log10_temperature", shift="log10_Tref")
ShiftedRTPoly = create_Poly("temperature", shift="Tref", reciprocal=True)
TPiecewise = create_Piecewise("temperature", nan_fallback=True)



# -*- coding: utf-8 -*-

import warnings

try:
    from numpy import any as _any
except ImportError:

    def _any(arg):
        if arg is True:
            return True
        if arg is False:
            return False
        return any(arg)


def water_density(T=None, T0=None, units=None, a=None, just_return_a=False, warn=True):
    """
    Density of water (kg/m3) as function of temperature (K)
    according to VSMOW model between 0 and 40 degree Celsius.
    Fitted using Thiesen's equation.

    Parameters
    ----------
    T : float
        Temperature (in Kelvin) (default: 298.15).
    T0 : float
        Value of T for 0 degree Celsius (default: 273.15).
    units : object (optional)
        Object with attributes: Kelvin, meter, kilogram.
    a : array_like (optional)
        5 parameters to the equation.
    just_return_a : bool (optional, default: False)
        Do not compute rho, just return the parameters ``a``.
    warn : bool (default: True)
        Emit UserWarning when outside temperature range.

    Returns
    -------
    Density of water (float of kg/m3 if T is float and units is None)

    Examples
    --------
    >>> print('%.2f' % water_density(277.13))
    999.97

    References
    ----------
    TANAKA M., GIRARD G., DAVIS R., PEUTO A. and BIGNELL N.,
        "Recommended table for the density of water between 0 °C and 40 °C
        based on recent experimental reports",
        Metrologia, 2001, 38, 301-309.
        http://iopscience.iop.org/article/10.1088/0026-1394/38/4/3
        doi:10.1088/0026-1394/38/4/3
    """
    if units is None:
        K = 1
        m = 1
        kg = 1
    else:
        K = units.Kelvin
        m = units.meter
        kg = units.kilogram
    if T is None:
        T = 298.15 * K
    m3 = m ** 3
    if a is None:
        a = (
            -3.983035 * K,  # C
            301.797 * K,  # C
            522528.9 * K * K,  # C**2
            69.34881 * K,  # C
            999.974950 * kg / m3,
        )
    if just_return_a:
        return a
    if T0 is None:
        T0 = 273.15 * K
    t = T - T0
    if warn and (_any(t < 0 * K) or _any(t > 40 * K)):
        warnings.warn("Temperature is outside range (0-40 degC)")
    return a[4] * (1 - ((t + a[0]) ** 2 * (t + a[1])) / (a[2] * (t + a[3])))


# bibtex format (generated at doi2bib.org):
reference = {
    "doi": "10.1088/0026-1394/38/4/3",
    "url": "http://dx.doi.org/10.1088/0026-1394/38/4/3",
    "year ": 2001,
    "month": "aug",
    "publisher": "{IOP} Publishing",
    "volume": 38,
    "number": 4,
    "pages": (301, 309),
    "author": "M Tanaka and G Girard and R Davis and A Peuto and N Bignell",
    "title": "Recommended table for the density of water between 0 ~C and 40 ~C based on recent experimental reports",
    "journal": "Metrologia",
}



from .numbers import number_to_scientific_latex
from .string import StrPrinter
from ..units import _latex_from_dimensionality


class LatexPrinter(StrPrinter):

    _default_settings = dict(
        StrPrinter._default_settings,
        repr_name="latex",
        Equilibrium_arrow=r"\rightleftharpoons",
        Reaction_arrow=r"\rightarrow",
        magnitude_fmt=number_to_scientific_latex,
        unit_fmt=_latex_from_dimensionality,
    )

    def _print_Substance(self, substance, **kwargs):
        return substance.latex_name or substance.name


def latex(obj, **settings):
    return LatexPrinter(settings).doprint(obj)



import math
from chempy import Reaction
from chempy.units import allclose, default_units as u
from ..testing import requires
from ..rendering import eval_template
from ..parsing import get_parsing_context
from chempy.units import units_library


@requires(units_library)
def test_eval_template():
    rendered = eval_template("${2*pi*arg*m**2}", arg=1 / math.pi)
    val = eval(rendered, get_parsing_context())
    assert allclose(val, 2 * u.m ** 2)


@requires(units_library)
def test_eval_template__Reaction():
    rendered = eval_template("2 OH -> H2O2; ${6*pi*arg}/M/s", arg=1 / math.pi)
    assert allclose(
        Reaction.from_string(rendered).param,
        Reaction.from_string("2 OH -> H2O2; 6.0/M/s").param,
    )



# -*- coding: utf-8 -*-

from ..util.testing import requires

try:
    import numpy as np
except ImportError:
    np, NumSysLin, NumSysLog = [None] * 3
else:
    from scipy.optimize import fsolve

from .._equilibrium import equilibrium_residual, solve_equilibrium, _get_rc_interval
from .._util import prodpow


@requires("numpy")
def test_equilibrium_residual():
    c0 = np.array((13.0, 11, 17))
    stoich = np.array((-2, 3, -4))
    K = 3.14
    assert (
        abs(
            equilibrium_residual(0.1, c0, stoich, K)
            - (K - (13 - 0.2) ** -2 * (11 + 0.3) ** 3 * (17 - 0.4) ** -4)
        )
        < 1e-14
    )


@requires("numpy")
def test_get__rc_interval():
    c = np.array((13.0, 11, 17))
    stoich = np.array((-2, 3, -4))
    limits = _get_rc_interval(stoich, c)
    lower = -11 / 3.0
    upper = 17.0 / 4
    assert abs(limits[0] - lower) < 1e-14
    assert abs(limits[1] - upper) < 1e-14


@requires("numpy")
def test_solve_equilibrium_1():
    c = np.array((13.0, 11, 17))
    stoich = np.array((-2, 3, -4))
    K = 3.14

    def f(x):
        return (13 - 2 * x) ** -2 * (11 + 3 * x) ** 3 * (17 - 4 * x) ** -4 - K

    assert np.allclose(solve_equilibrium(c, stoich, K), c + stoich * fsolve(f, 3.48))


@requires("numpy")
def test_solve_equilibrium_2():
    c = np.array([1.7e-03, 3.0e06, 3.0e06, 9.7e07, 5.55e09])
    stoich = (1, 1, 0, 0, -1)
    K = 55 * 1e-6

    def f(x):
        return prodpow(c + x * stoich, stoich) - K

    solution = solve_equilibrium(c, stoich, K)
    assert np.allclose(solution, c + stoich * fsolve(f, 0.1))



# -*- coding: utf-8 -*-
"""
Contains functions for the `Arrhenius equation
<https://en.wikipedia.org/wiki/Arrhenius_equation>`_
(:func:`arrhenius_equation`) and a convenience fitting routine
(:func:`fit_arrhenius_equation`).
"""

from .._util import get_backend
from ..util.regression import least_squares
from ..util.pyutil import defaultnamedtuple
from ..units import default_constants, default_units, format_string, patched_numpy

try:
    import numpy as np
except ImportError:
    np = None


def _fit_linearized(backtransfm, lin_x, lin_y, lin_yerr):
    if len(lin_x) != len(lin_y):
        raise ValueError("k and T needs to be of equal length.")
    if lin_yerr is not None:
        if len(lin_yerr) != len(lin_y):
            raise ValueError("kerr and T needs to be of equal length.")
    lin_p, lin_vcv, lin_r2 = least_squares(lin_x, lin_y, lin_yerr)
    return [cb(lin_p) for cb in backtransfm]


def _fit(T, k, kerr, func, lin_x, lin_y, backtransfm, linearized=False):
    _lin_y = lin_y(T, k)
    if kerr is None:
        lin_yerr = 1
    else:
        lin_yerr = (
            abs(lin_y(k - kerr, T) - _lin_y) + abs(lin_y(k + kerr, T) - _lin_y)
        ) / 2

    lopt = _fit_linearized(backtransfm, lin_x(T, k), _lin_y, lin_yerr)
    if linearized:
        return lopt
    from scipy.optimize import curve_fit

    popt, pcov = curve_fit(func, T, k, lopt, kerr)
    return popt, pcov


def _get_R(constants=None, units=None):
    if constants is None:
        R = 8.314472
        if units is not None:
            J = units.Joule
            K = units.Kelvin
            mol = units.mol
            R *= J / mol / K
    else:
        R = constants.molar_gas_constant.simplified
    return R


def arrhenius_equation(A, Ea, T, constants=None, units=None, backend=None):
    """
    Returns the rate coefficient according to the Arrhenius equation

    Parameters
    ----------
    A: float with unit
        frequency factor
    Ea: float with unit
        activation energy
    T: float with unit
        temperature
    constants: object (optional, default: None)
        if None:
            T assumed to be in Kelvin, Ea in J/(K mol)
        else:
            attributes accessed: molar_gas_constant
            Tip: pass quantities.constants
    units: object (optional, default: None)
        attributes accessed: Joule, Kelvin and mol
    backend: module (optional)
        module with "exp", default: numpy, math

    """
    be = get_backend(backend)
    R = _get_R(constants, units)
    try:
        RT = (R * T).rescale(Ea.dimensionality)
    except AttributeError:
        RT = R * T
    return A * be.exp(-Ea / RT)


def fit_arrhenius_equation(
    T, k, kerr=None, linearized=False, constants=None, units=None
):
    """Curve fitting of the Arrhenius equation to data points

    Parameters
    ----------
    T : float
    k : array_like
    kerr : array_like (optional)
    linearized : bool

    """
    return _fit(
        T,
        k,
        kerr,
        arrhenius_equation,
        lambda T, k: 1 / T,
        lambda T, k: np.log(k),
        [lambda p: np.exp(p[0]), lambda p: -p[1] * _get_R(constants, units)],
        linearized=linearized,
    )


def _fit_arrhenius_equation(T, k, kerr=None, linearized=False):
    """Curve fitting of the Arrhenius equation to data points

    Parameters
    ----------
    k : array_like
    T : float
    kerr : array_like (optional)
    linearized : bool

    """
    if len(k) != len(T):
        raise ValueError("k and T needs to be of equal length.")
    from math import exp
    import numpy as np

    p = np.polyfit(1 / T, np.log(k), 1)
    R = _get_R(constants=None, units=None)
    Ea = -R * p[0]
    A = exp(p[1])
    if linearized:
        return A, Ea
    from scipy.optimize import curve_fit

    if kerr is None:
        weights = None
    else:
        weights = 1 / kerr ** 2
    popt, pcov = curve_fit(arrhenius_equation, T, k, [A, Ea], weights)
    return popt, pcov


class ArrheniusParam(defaultnamedtuple("ArrheniusParam", "A Ea ref", [None])):
    """Kinetic data in the form of an Arrhenius parameterisation

    Parameters
    ----------
    Ea: float
        activation energy
    A: float
        preexponential prefactor (Arrhenius type eq.)
    ref: object (default: None)
        arbitrary reference (e.g. string representing citation key)

    Examples
    --------
    >>> k = ArrheniusParam(1e13, 40e3)
    >>> '%.5g' % k(298.15)
    '9.8245e+05'

    """

    def html(self, fmt):
        return "%s exp((%s)/(RT))" % (fmt(self.A), fmt(self.Ea))

    def unicode(self, fmt):
        return "%s exp((%s)/(RT))" % (fmt(self.A), fmt(self.Ea))

    @classmethod
    def from_rateconst_at_T(
        cls, Ea, T_k, backend=None, constants=None, units=None, **kwargs
    ):
        """Constructs an instance from a known rate constant at a given temperature.

        Parameters
        ----------
        Ea : float
            Activation energy.
        T_k : tuple of two floats
            Temperature & rate constant.

        """
        # k = A*exp(-Ea/R/T)
        # A = k*exp(Ea/R/T)
        T, k = T_k
        R = _get_R(constants, units)
        if backend is None:
            from chempy.units import patched_numpy as backend
        return cls(k * backend.exp(Ea / R / T), Ea, **kwargs)

    @classmethod
    def from_fit_of_data(cls, T, k, kerr=None, **kwargs):
        args, vcv = fit_arrhenius_equation(T, k, kerr)
        return cls(*args, **kwargs)

    def __call__(self, T, constants=None, units=None, backend=None):
        """Evaluates the arrhenius equation for a specified state

        Parameters
        ----------
        T: float
        constants: module (optional)
        units: module (optional)
        backend: module (default: math)

        See also
        --------
        chempy.arrhenius.arrhenius_equation : the function called here.

        """
        return arrhenius_equation(
            self.A, self.Ea, T, constants=constants, units=units, backend=backend
        )

    def Ea_over_R(self, constants, units, backend=None):
        return self.Ea / _get_R(constants, units)

    def as_RateExpr(self, unique_keys=None, constants=None, units=None, backend=None):
        from .rates import Arrhenius, MassAction

        args = [self.A, self.Ea_over_R(constants, units)]
        return MassAction(Arrhenius(args, unique_keys))

    def format(self, precision, tex=False):
        try:
            str_A, str_A_unit = format_string(self.A, precision, tex)
            str_Ea, str_Ea_unit = format_string(self.Ea, precision, tex)
        except Exception:
            str_A, str_A_unit = precision.format(self.A), "-"
            str_Ea, str_Ea_unit = precision.format(self.Ea), "-"
        return (str_A, str_A_unit), (str_Ea, str_Ea_unit)

    def equation_as_string(self, precision, tex=False):
        (str_A, str_A_unit), (str_Ea, str_Ea_unit) = self.format(precision, tex)
        if tex:
            return (
                r"{}\exp \left(-\frac{{{}}}{{RT}} \right)".format(
                    str_A, str_Ea + " " + str_Ea_unit
                ),
                str_A_unit,
            )
        else:
            return (
                "{}*exp(-{}/(R*T))".format(str_A, str_Ea + " " + str_Ea_unit),
                str_A_unit,
            )

    def __str__(self):
        return " ".join(self.equation_as_string("%.5g"))


class ArrheniusParamWithUnits(ArrheniusParam):
    @classmethod
    def from_rateconst_at_T(cls, *args, **kwargs):
        if "constants" not in kwargs:
            kwargs["constants"] = default_constants
        if "units" not in kwargs:
            kwargs["units"] = default_units
        if "backend" not in kwargs:
            kwargs["backend"] = patched_numpy
        return super(ArrheniusParamWithUnits, cls).from_rateconst_at_T(*args, **kwargs)

    def __call__(
        self, state, constants=default_constants, units=default_units, backend=None
    ):
        """See :func:`chempy.arrhenius.arrhenius_equation`."""
        return super(ArrheniusParamWithUnits, self).__call__(
            state, constants, units, backend
        )

    def as_RateExpr(
        self, unique_keys=None, constants=default_constants, units=default_units
    ):
        return super(ArrheniusParamWithUnits, self).as_RateExpr(
            unique_keys, constants, units
        )



# -*- coding: utf-8 -*-

import math

import pytest

from chempy import Reaction, ReactionSystem, Substance
from chempy.chemistry import Equilibrium
from chempy.units import (
    allclose,
    Backend,
    to_unitless,
    units_library,
    patched_numpy,
    default_constants,
    simplified,
    default_units as u,
)
from chempy.util._expr import Expr
from chempy.util.parsing import parsing_library
from chempy.util.testing import requires
from ..rates import RateExpr, MassAction, Arrhenius, Radiolytic, mk_Radiolytic, Eyring


class SpecialFraction(RateExpr):
    """
    Example from Atkins, De Paula, Physical Chemistry
    H2 + Br2 -> 2HBr
    ½ dHBr/dt = k[H2][Br2]**(3/2) / ([Br2] + k'[HBr])
    """

    def __call__(self, variables, backend=math, reaction=None):
        two = 2 * backend.pi ** 0
        k = self.arg(variables, 0)
        kp = self.arg(variables, 1)
        H2, Br2, HBr = variables["H2"], variables["Br2"], variables["HBr"]
        return k * H2 * Br2 ** (3 / two) / (Br2 + kp * HBr)


def _get_SpecialFraction_rsys(k, kprime):
    ratex = SpecialFraction([k, kprime], ["k_HBr", "kprime_HBr"])
    rxn = Reaction({"H2": 1, "Br2": 1}, {"HBr": 2}, ratex)
    return ReactionSystem([rxn], "H2 Br2 HBr")


@requires("numpy")
def test_specialfraction_rateexpr():
    rsys = _get_SpecialFraction_rsys(11, 13)
    r = rsys.rxns[0]

    conc = {"H2": 2, "Br2": 3, "HBr": 5}

    def _check(k, kprime, c):
        ref = k * c["H2"] * c["Br2"] ** 1.5 / (c["Br2"] + kprime * c["HBr"])
        assert abs(r.rate_expr()(c) - ref) < 1e-15

    _check(11, 13, conc)


def test_RateExpr__subclass_from_callback():
    SF = RateExpr.subclass_from_callback(
        lambda v, a, backend: a[0]
        * v["H2"]
        * v["Br2"] ** (3 / 2)
        / (v["Br2"] + a[1] * v["HBr"])
    )
    ratex = SF([11, 13], ["k_HBr", "kprime_HBr"])
    r = Reaction({"H2": 1, "Br2": 1}, {"HBr": 2}, ratex)

    res1 = r.rate_expr()({"H2": 5, "Br2": 7, "HBr": 15})
    ref1 = 11 * 5 * 7 ** 1.5 / (7 + 13 * 15)
    assert abs((res1 - ref1) / ref1) < 1e-14

    res2 = r.rate_expr()({"H2": 5, "Br2": 7, "HBr": 15, "k_HBr": 23, "kprime_HBr": 42})
    ref2 = 23 * 5 * 7 ** 1.5 / (7 + 42 * 15)
    assert abs((res2 - ref2) / ref2) < 1e-14


def test_MassAction():
    ma = MassAction([3.14], ["my_rate"])
    assert ma.nargs == 1
    r = Reaction({"A": 2, "B": 1}, {"C": 1}, ma, {"B": 1})
    arg1 = {"A": 11, "B": 13, "C": 17}
    arg2 = {"A": 11, "B": 13, "C": 17, "my_rate": 2.72}
    res1 = r.rate_expr()(arg1, reaction=r)
    res2 = r.rate_expr()(arg2, reaction=r)
    ref1 = 3.14 * 13 * 11 ** 2
    ref2 = 2.72 * 13 * 11 ** 2
    assert abs(res1 - ref1) < 1e-14
    assert abs(res2 - ref2) < 1e-12
    rat1 = r.rate(arg1)
    rat2 = r.rate(arg2)
    for key, coeff in [("A", -2), ("B", -2), ("C", 1)]:
        assert abs(rat1[key] - ref1 * coeff) < 2e-12
        assert abs(rat2[key] - ref2 * coeff) < 2e-12


@requires(units_library)
def test_MassAction__expression():
    class GibbsExpr(Expr):
        parameter_keys = ("temperature",)
        argument_names = tuple("dS_over_R dCp_over_R dH_over_R Tref".split())

        def __call__(self, variables, backend=patched_numpy, **kwargs):
            am = dict(
                zip(
                    self.argument_names,
                    map(simplified, self.all_args(variables, backend=backend)),
                )
            )
            (T,) = self.all_params(variables, backend=backend)
            return (
                backend.exp(am["dS_over_R"])
                * (T / am["Tref"]) ** am["dCp_over_R"]
                * backend.exp(-am["dH_over_R"] / T)
            )

    GeNH3 = GibbsExpr(
        dict(
            dS_over_R=18.8
            * u.cal
            / u.K
            / u.mol
            / default_constants.molar_gas_constant,  # NOQA
            dCp_over_R=52
            * u.cal
            / u.K
            / u.mol
            / default_constants.molar_gas_constant,  # NOQA
            dH_over_R=-0.87e3
            * u.cal
            / u.mol
            / default_constants.molar_gas_constant,  # NOQA
            Tref=298.15 * u.K,  # NOQA
        )
    )
    reac_prod = {"NH3": 1, "H2O": 1}, {"NH4+": 1, "OH-": 1}
    Equilibrium(*reac_prod, GeNH3).check_consistent_units(throw=True)
    Equilibrium(*reac_prod[::-1], 1 / GeNH3).check_consistent_units(throw=True)

    Ea = 40e3 * u.J / u.mol
    R = default_constants.molar_gas_constant
    A, Ea_over_R = 1.2e11 / u.molar ** 2 / u.second, Ea / R
    arrh = Arrhenius([A, Ea_over_R])
    ama = MassAction(arrh)

    ma_mul_expr = ama * GeNH3
    ma_div_expr = ama / GeNH3
    expr_mul_ma = GeNH3 * ama
    expr_div_ma = GeNH3 / ama
    assert all(
        isinstance(expr, MassAction)
        for expr in [ma_mul_expr, ma_div_expr, expr_mul_ma, expr_div_ma]
    )

    Reaction(*reac_prod, ama).check_consistent_units(throw=True)
    Reaction(*reac_prod[::-1], 42 * ma_div_expr).check_consistent_units(throw=True)
    Reaction(*reac_prod[::-1], 42 / u.M / u.s / GeNH3).check_consistent_units(
        throw=True
    )

    varbls = {"temperature": 298.15 * u.K}
    r_ama, r_GeNH3 = ama.rate_coeff(varbls), GeNH3(varbls)
    assert allclose(ma_mul_expr.rate_coeff(varbls), r_ama * r_GeNH3)
    assert allclose(ma_div_expr.rate_coeff(varbls), r_ama / r_GeNH3)
    assert allclose(expr_mul_ma.rate_coeff(varbls), r_GeNH3 * r_ama)
    assert allclose(expr_div_ma.rate_coeff(varbls), r_GeNH3 / r_ama)


def test_Expr__from_callback():
    def rate_coeff(args, T, reaction, backend=math):
        return args[0] * backend.exp(args[1] / T)

    RateExpr = Expr.from_callback(rate_coeff, parameter_keys=("Tem",), nargs=2)
    k1 = RateExpr([2.1e10, -5132.2])
    rxn = Reaction({"H2": 2, "O2": 1}, {"H2O": 2}, MassAction(k1))
    ma = rxn.rate_expr()
    res = ma({"Tem": 273.15, "H2": 7, "O2": 13}, reaction=rxn)
    ref = 7 * 7 * 13 * 2.1e10 * math.exp(-5132.2 / 273.15)
    assert abs((res - ref) / ref) < 1e-14


def test_MassAction__subclass_from_callback():
    def rate_coeff(variables, all_args, reaction, backend):
        return all_args[0] * backend.exp(all_args[1] / variables["temperature"])

    CustomMassAction = MassAction.subclass_from_callback(
        rate_coeff, cls_attrs=dict(parameter_keys=("temperature",), nargs=2)
    )
    k1 = CustomMassAction([2.1e10, -5132.2])
    rxn = Reaction({"H2": 2, "O2": 1}, {"H2O": 2}, k1)
    cma = rxn.rate_expr()
    res = cma({"temperature": 273.15, "H2": 7, "O2": 13}, reaction=rxn)
    ref = 7 * 7 * 13 * 2.1e10 * math.exp(-5132.2 / 273.15)
    assert abs((res - ref) / ref) < 1e-14


@requires(units_library)
def test_MassAction__subclass_from_callback__units():
    def rate_coeff(variables, all_args, backend, **kwargs):
        return all_args[0] * backend.exp(all_args[1] / variables["temperature"])

    CustomMassAction = MassAction.subclass_from_callback(
        rate_coeff, cls_attrs=dict(parameter_keys=("temperature",), nargs=2)
    )
    k1 = CustomMassAction([2.1e10 / u.molar ** 2 / u.second, -5132.2 * u.kelvin])
    rxn = Reaction({"H2": 2, "O2": 1}, {"H2O": 2}, k1)
    variables = {
        "temperature": 491.67 * u.rankine,
        "H2": 7000 * u.mol / u.metre ** 3,
        "O2": 13 * u.molar,
    }
    cma = rxn.rate_expr()
    res = cma(variables, backend=Backend(), reaction=rxn)
    ref = 7 * 7 * 13 * 2.1e10 * math.exp(-5132.2 / 273.15) * u.molar / u.second
    assert allclose(res, ref)


def test_ArrheniusMassAction():
    A, Ea_over_R = 1.2e11, 40e3 / 8.3145
    ama = MassAction(Arrhenius([A, Ea_over_R]))
    r = Reaction({"A": 2, "B": 1}, {"C": 1}, ama, {"B": 1})
    T_ = "temperature"

    def ref(v):
        return 1.2e11 * math.exp(-Ea_over_R / v[T_]) * v["B"] * v["A"] ** 2

    ma = r.rate_expr()
    for params in [(11.0, 13.0, 17.0, 311.2), (12, 8, 5, 270)]:
        var = dict(zip(["A", "B", "C", T_], params))
        ref_var = ref(var)
        assert abs((ma(var, reaction=r) - ref_var) / ref_var) < 1e-14

    with pytest.raises(ValueError):
        Arrhenius([A, Ea_over_R, 1, A])

    # assert ama.as_mass_action({T_: 273.15}).args[0] == A*math.exp(-Ea_over_R/273.15)


@requires(units_library)
def test_specialfraction_rateexpr__units():
    _k, _kprime = 3.5 * u.s ** -1 * u.molar ** -0.5, 9.2
    rsys = _get_SpecialFraction_rsys(_k, _kprime)
    r = rsys.rxns[0]

    conc = {"H2": 2 * u.molar, "Br2": 3000 * u.mol / u.m ** 3, "HBr": 5 * u.molar}
    ma = r.rate_expr()

    def _check(k, kprime, c):
        ref = k * c["H2"] * c["Br2"] ** 1.5 / (c["Br2"] + kprime * c["HBr"])
        assert abs(ma(c, reaction=r) - ref) < 1e-15 * u.molar / u.second

    _check(_k, _kprime, conc)
    alt_k, alt_kprime = 2 * u.s ** -1 * u.molar ** -0.5, 5
    _check(
        alt_k,
        alt_kprime,
        dict(list(conc.items()) + [("k_HBr", alt_k), ("kprime_HBr", alt_kprime)]),
    )


@requires(units_library, "numpy")
@pytest.mark.parametrize("R_from_constants", [False, True])
def test_ArrheniusMassAction__units(R_from_constants):
    import numpy as np

    Ea = 40e3 * u.J / u.mol
    R = (
        default_constants.molar_gas_constant
        if R_from_constants
        else (8.3145 * u.J / u.mol / u.K)
    )
    A, Ea_over_R = 1.2e11 / u.molar ** 2 / u.second, Ea / R
    ref1 = A * np.exp(-to_unitless(Ea_over_R / (290 * u.K)))
    arrh = Arrhenius([A, Ea_over_R])
    assert allclose(arrh({"temperature": 290 * u.K}), ref1)
    ama = MassAction(arrh)
    r = Reaction({"A": 2, "B": 1}, {"C": 1}, ama, {"B": 1})
    T_ = "temperature"

    def ref(v):
        return (
            1.2e11
            / u.molar ** 2
            / u.second
            * math.exp(-Ea_over_R.simplified / v[T_])
            * v["B"]
            * v["A"] ** 2
        )

    ma = r.rate_expr()

    for params in [
        (11.0 * u.molar, 13.0 * u.molar, 17.0 * u.molar, 311.2 * u.kelvin),
        (12 * u.molar, 8 * u.molar, 5 * u.molar, 270 * u.kelvin),
    ]:
        var = dict(zip(["A", "B", "C", T_], params))
        ref_val = ref(var)
        assert abs((ma(var, reaction=r) - ref_val) / ref_val) < 1e-14


def test_Radiolytic():
    r = Radiolytic([2.1e-7])
    res = r({"doserate": 0.15, "density": 0.998})
    assert abs(res - 0.15 * 0.998 * 2.1e-7) < 1e-15


@requires(parsing_library)
def test_Radiolytic__parsing():
    rxn = Reaction.from_string(
        "-> H + OH; Radiolytic({'radiolytic_yield': 2.1e-7})", None
    )
    res = rxn.rate({"doserate": 0.15, "density": 0.998})
    ref = 0.15 * 0.998 * 2.1e-7
    assert abs((res["H"] - ref) / ref) < 1e-15
    assert abs((res["OH"] - ref) / ref) < 1e-15
    (gval,) = rxn.rate_expr().g_values({}).values()
    assert abs(gval - 2.1e-7) < 1e-15


@requires(parsing_library, units_library)
def test_Radiolytic__parsing__units():
    rxn = Reaction.from_string(
        "-> H + OH; Radiolytic({'radiolytic_yield': 2.1e-7*mol/J})", None
    )
    assert rxn.reac == {}
    assert rxn.prod == {"H": 1, "OH": 1}
    res = rxn.rate({"doserate": 0.15 * u.gray / u.s, "density": 0.998 * u.kg / u.dm3})
    ref = 0.15 * 0.998 * 2.1e-7 * u.molar / u.second
    assert abs((res["H"] - ref) / ref) < 1e-15
    assert abs((res["OH"] - ref) / ref) < 1e-15


@requires(units_library)
def test_Radiolytic__units():
    def _check(r):
        res = r(
            {
                "doserate": 0.15 * u.gray / u.second,
                "density": 0.998 * u.kg / u.decimetre ** 3,
            }
        )
        ref = 0.15 * 0.998 * 2.1e-7 * u.molar / u.second
        assert abs(to_unitless((res - ref) / ref)) < 1e-15

    _check(Radiolytic([2.1e-7 * u.mol / u.joule]))
    _check(Radiolytic([2.0261921896167396 * u.per100eV]))


@requires(units_library)
def test_Radioyltic__Reaction_html():
    rate = Radiolytic([2.1 * u.per100eV])
    rxn = Reaction({}, {"H": 1}, rate)
    H = Substance.from_formula("H")
    html = rxn.html({"H": H}, with_param=True)
    assert html == " &rarr; H&#59; %s" % str(rate)


def test_mk_Radiolytic():
    R1 = mk_Radiolytic()
    R2 = mk_Radiolytic()
    assert R1 is R2

    RABG = mk_Radiolytic("alpha", "beta", "gamma")
    rxn = Reaction({}, {"H": 2}, RABG([3, 5, 7], "ya yb yg".split()))
    rat = rxn.rate(
        {
            "doserate_alpha": 11,
            "doserate_beta": 13,
            "doserate_gamma": 17,
            "density": 0.7,
        }
    )
    assert abs(rat["H"] - 0.7 * 2 * (3 * 11 + 5 * 13 + 7 * 17)) < 1e-13
    assert RABG.parameter_keys == (
        "density",
        "doserate_alpha",
        "doserate_beta",
        "doserate_gamma",
    )
    assert RABG.argument_names == tuple(
        "radiolytic_yield_%s" % k for k in "alpha beta gamma".split()
    )
    assert rxn.param.unique_keys == ("ya", "yb", "yg")
    rat2 = rxn.rate(
        {
            "doserate_alpha": 11,
            "doserate_beta": 13,
            "doserate_gamma": 17,
            "density": 0.7,
            "ya": 23,
            "yb": 29,
            "yg": 31,
        }
    )
    assert abs(rat2["H"] - 0.7 * 2 * (23 * 11 + 29 * 13 + 31 * 17)) < 1e-13


class Eyring4(Expr):
    nargs = 4

    def __call__(self, variables, reaction, backend=math):
        Sact_fact, Hact_exp, Sref_fact, Href_exp = self.all_args(
            variables, backend=backend
        )
        T = variables["temperature"]
        return T * Sact_fact / Sref_fact * backend.exp((Href_exp - Hact_exp) / T)


def test_EyringMassAction():
    args = kB_h_times_exp_dS_R, dH_over_R, c0 = 1.2e11 / 273.15, 40e3 / 8, 1
    ama = MassAction(Eyring(args, ("Sfreq", "Hact")))
    rxn1 = Reaction({"A": 2, "B": 1}, {"C": 1}, ama, {"B": 1})
    T_ = "temperature"

    def ref(v):
        return (
            v.get("Sfreq", 1.2e11 / 273.15)
            * v[T_]
            * math.exp(-v.get("Hact", 40e3 / 8) / v[T_])
            * v["B"]
            * v["A"] ** 2
        )

    for params in [(11.0, 13.0, 17.0, 311.2), (12, 8, 5, 270)]:
        var = dict(zip(["A", "B", "C", T_], params))
        ref_val = ref(var)
        assert abs((ama(var, reaction=rxn1) - ref_val) / ref_val) < 1e-14

    with pytest.raises(ValueError):
        MassAction(Eyring([1, 1, 1, 1, 1]))

    # assert ama.as_mass_action({T_: 273.15}).args[0] == 1.2e11*math.exp(-40e3/8/273.15)

    ama2 = MassAction(
        Eyring4([1.2e11 / 273, 40e3 / 8, 1.2, 1e3], ("Sfreq", "Hact", "Sref", "Href"))
    )
    rxn2 = Reaction({"C": 1}, {"A": 2, "B": 2}, ama2)
    var2 = {"C": 29, "temperature": 273}

    def ref2(var):
        return (
            var["C"]
            * var.get("temperature", 273)
            * var.get("Sfreq", 1.2e11 / 273)
            / var.get("Sref", 1.2)
            * math.exp(
                (var.get("Href", 1e3) - var.get("Hact", 5e3))
                / var.get("temperature", 273)
            )
        )

    r2 = ref2(var2)
    assert abs((ama2(var2, reaction=rxn2) - r2) / r2) < 1e-14

    rsys = ReactionSystem([rxn1, rxn2])
    var3 = {"A": 11, "B": 13, "C": 17, "temperature": 298, "Sfreq": 1.2e11 / 298}
    rates = rsys.rates(var3)
    rf3 = ref(var3)
    rb3 = ref2(var3)
    ref_rates = {"A": 2 * (rb3 - rf3), "B": 2 * (rb3 - rf3), "C": rf3 - rb3}
    for k, v in ref_rates.items():
        assert abs((rates[k] - v) / v) < 1e-14



"""
Interactive kinetics app with sliders (with units).
Start by running:
    $ bokeh serve interactive.py
Add --show argument or navigate to:
    http://localhost:5006/interactive
"""
from collections import defaultdict
import sys

from chempy.util.bkh import integration_with_sliders
from chempy.units import SI_base_registry, default_units as u

from bokeh_interactive_arrhenius import get_rsys


if __name__.startswith("bk_"):
    from bokeh.io import curdoc

    Af, Ab, Ea, Er = (
        1e16 / u.molar / u.s,
        1.5e15 / u.s,
        72e3 * u.J / u.mol,
        -12e3 * u.J / u.mol,
    )
    curdoc().add_root(
        integration_with_sliders(
            get_rsys(Af, Ab, Ea, Er),
            tend=3 * u.s,
            c0=defaultdict(
                lambda: 0 * u.molar, {"Fe+3": 3e-3 * u.molar, "SCN-": 1.5e-3 * u.molar}
            ),
            parameters={"temperature": 298.15 * u.K},
            slider_kwargs={
                "temperature": dict(
                    start=273.15 * u.K, end=313.15 * u.K, step=0.05 * u.K
                )
            },
            get_odesys_kw=dict(
                unit_registry=SI_base_registry,
                output_conc_unit=u.molar,
                output_time_unit=u.second,
            ),
        )
    )
elif __name__ == "__main__":
    import warnings

    warnings.warn("Run using 'bokeh serve %s'" % __file__)
    sys.exit(1)



# -*- coding: utf-8 -*-
"""
Convenience functions for representing reaction systems as graphs.
"""
import os
import subprocess
import shutil
import tempfile


def rsys2dot(
    rsys,
    tex=False,
    rprefix="r",
    rref0=1,
    nodeparams='[label="{}",shape=diamond]',
    colors=("maroon", "darkgreen"),
    penwidths=None,
    include_inactive=True,
):
    """
    Returns list of lines of DOT (graph description language)
    formatted graph.

    Parameters
    ==========
    rsys: ReactionSystem
    tex: bool (default False)
        If set True, output will be LaTeX formatted
    (Substance need to have latex_name attribute set)
    rprefix: string
        Reaction enumeration prefix, default: r
    rref0: integer
        Reaction enumeration initial counter value, default: 1
    nodeparams: string
        DOT formatted param list, default: [label={} shape=diamond]

    Returns
    =======
    list of lines of DOT representation of the graph representation.

    """
    lines = ['digraph "' + str(rsys.name) + '" {\n']
    ind = "  "  # indentation
    if penwidths is None:
        penwidths = [1.0] * rsys.nr

    categories = rsys.categorize_substances(checks=())

    def add_substance(key):
        fc = "black"
        if key in categories["depleted"]:
            fc = colors[0]
        if key in categories["accumulated"]:
            fc = colors[1]
        label = ("$%s$" if tex else "%s") % getattr(
            rsys.substances[key], "latex_name" if tex else "name"
        )
        lines.append(
            ind
            + '"{key}" [fontcolor={fc} label="{lbl}"];\n'.format(
                key=key, fc=fc, lbl=label
            )
        )

    for sk in rsys.substances:
        add_substance(sk)

    def add_vertex(key, num, reac, penwidth):
        snum = str(num) if num > 1 else ""
        fmt = ",".join(
            ['label="{}"'.format(snum)]
            + (["penwidth={}".format(penwidth)] if penwidth != 1 else [])
        )
        lines.append(
            ind
            + '"{}" -> "{}" [color={},fontcolor={},{}];\n'.format(
                *(
                    (key, rid, colors[0], colors[0], fmt)
                    if reac
                    else (rid, key, colors[1], colors[1], fmt)
                )
            )
        )

    if include_inactive:
        reac_stoichs = rsys.all_reac_stoichs()
        prod_stoichs = rsys.all_prod_stoichs()
    else:
        reac_stoichs = rsys.active_reac_stoichs()
        prod_stoichs = rsys.active_prod_stoichs()

    for ri, rxn in enumerate(rsys.rxns):
        rid = rprefix + str(ri + rref0)
        lines.append(ind + "{")
        lines.append(ind * 2 + "node " + nodeparams.format(rxn.name or rid))
        lines.append(ind * 2 + rid)
        lines.append(ind + "}\n")
        for idx, key in enumerate(rsys.substances):
            num = reac_stoichs[ri, idx]
            if num == 0:
                continue
            add_vertex(key, num, True, penwidths[ri])
        for idx, key in enumerate(rsys.substances):
            num = prod_stoichs[ri, idx]
            if num == 0:
                continue
            add_vertex(key, num, False, penwidths[ri])
    lines.append("}\n")
    return lines


def rsys2graph(rsys, fname, output_dir=None, prog=None, save=False, **kwargs):
    """
    Convenience function to call `rsys2dot` and write output to file
    and render the graph

    Parameters
    ----------
    rsys : ReactionSystem
    fname : str
        filename
    output_dir : str (optional)
        path to directory (default: temporary directory)
    prog : str (optional)
        default: 'dot'
    save : bool
        removes temporary directory if False, default: False
    \\*\\*kwargs :
        Keyword arguments passed along to py:func:`rsys2dot`.

    Returns
    -------
    str
        Outpath

    Examples
    --------
    >>> rsys2graph(rsys, sbstncs, '/tmp/out.png')  # doctest: +SKIP

    """

    lines = rsys2dot(rsys, **kwargs)
    created_tempdir = False
    try:
        if output_dir is None:
            output_dir = tempfile.mkdtemp()
            created_tempdir = True
        basename, ext = os.path.splitext(os.path.basename(fname))
        outpath = os.path.join(output_dir, fname)
        dotpath = os.path.join(output_dir, basename + ".dot")
        with open(dotpath, "wt") as ofh:
            ofh.writelines(lines)
        if ext == ".tex":
            cmds = [prog or "dot2tex"]
        else:
            cmds = [prog or "dot", "-T" + outpath.split(".")[-1]]
        p = subprocess.Popen(cmds + [dotpath, "-o", outpath])
        retcode = p.wait()
        if retcode:
            fmtstr = "{}\n returned with exit status {}"
            raise RuntimeError(fmtstr.format(" ".join(cmds), retcode))
        return outpath
    finally:
        if save is True or save == "True":
            pass
        else:
            if save is False or save == "False":
                if created_tempdir:
                    shutil.rmtree(output_dir)
            else:
                # interpret save as path to copy pdf to.
                shutil.copy(outpath, save)



from .numbers import number_to_scientific_html
from .string import StrPrinter


def _html_clsname(key):
    return "chempy_" + key.replace("+", "plus").replace("-", "minus").replace(
        "(", "leftparen"
    ).replace(")", "rightparen")


_html_semicolon = "&#59; "


class HTMLPrinter(StrPrinter):

    printmethod_attr = "_html"
    _default_settings = dict(
        StrPrinter._default_settings,
        repr_name="html",
        Equilibrium_arrow="&harr;",
        Reaction_arrow="&rarr;",
        Reaction_param_separator=_html_semicolon,
        magnitude_fmt=number_to_scientific_html,
    )

    def _print_Substance(self, s, **kwargs):
        return s.html_name or s.name

    def _print_ReactionSystem(self, rsys, **kwargs):
        return (
            super(HTMLPrinter, self)
            ._print_ReactionSystem(rsys, **kwargs)
            .replace("\n", "<br>\n")
        )


def html(obj, **settings):
    return HTMLPrinter(settings).doprint(obj)


class CSSPrinter(HTMLPrinter):
    def _print_Substance(self, s, **kwargs):
        key = s.name
        name = s.html_name or key
        common_sty = "border-radius: 5pt; padding: 0pt 3pt 0pt 3pt;"
        colors = self._get("colors", **kwargs)
        if key in colors:
            style = "background-color:#%s; border: 1px solid #%s; %s" % (
                colors[key] + (common_sty,)
            )
        else:
            style = common_sty
        fmt = '<span class="%s" style="%s">%s</span>'
        return fmt % (_html_clsname(key), style, name)

    def _tr_id(self, rsys, i):
        return "chempy_%d_%d" % (id(rsys), i)

    def _print_ReactionSystem(self, rsys, **kwargs):
        sep = '</td><td style="text-align:left;">&nbsp;'
        around = (
            '</td><td style="text-align:center;">',
            '</td><td style="text-align:left;">',
        )
        # cf. https://github.com/jupyter/notebook/issues/2160#issuecomment-352216152
        row_template = '<tr class="%s"><td style="text-align:right;">%s</td></tr>'
        rows = [
            row_template % (self._tr_id(rsys, i), s)
            for i, s in enumerate(
                map(
                    lambda r: self._print(
                        r, Reaction_param_separator=sep, Reaction_around_arrow=around
                    ),
                    rsys.rxns,
                )
            )
        ]
        tab_template = '<table class="chempy_ReactionSystem chempy_%d">%s%s</table>'
        header = '<tr><th style="text-align:center;" colspan="5">%s</th></tr>' % (
            rsys.name or ""
        )
        return tab_template % (id(rsys), header, "\n\n".join(rows))


def css(obj, **settings):
    return CSSPrinter(settings).doprint(obj)



class _RxnTable(object):

    _rsys_meth = None

    def __init__(
        self,
        idx_rxn_pairs,
        substances,
        colors=None,
        missing=None,
        missing_color="eee8aa",
    ):
        self.idx_rxn_pairs = idx_rxn_pairs
        self.substances = substances
        self.colors = colors or {}
        self.missing = missing or []
        self.missing_color = missing_color

    @classmethod
    def from_ReactionSystem(cls, rsys, color_categories=True):
        idx_rxn_pairs, unconsidered_ri = getattr(rsys, cls._rsys_meth)()
        colors = rsys._category_colors() if color_categories else {}
        missing = [not rsys.substance_participation(sk) for sk in rsys.substances]
        return (
            cls(idx_rxn_pairs, rsys.substances, colors=colors, missing=missing),
            unconsidered_ri,
        )

    def _repr_html_(self):
        from .web import css

        return css(self, substances=self.substances, colors=self.colors)

    def _cell_label_html(self, printer, ori_idx, rxn):
        """Reaction formatting callback. (reaction index -> string)"""
        pretty = rxn.unicode(self.substances, with_param=True, with_name=False)
        return '<a title="%d: %s">%s</a>' % (
            ori_idx,
            pretty,
            printer._print(rxn.name or rxn.param),
        )

    def _cell_html(self, printer, A, ri, ci=None):
        args = []
        if ci is not None and ri > ci:
            r = "-"
        else:
            if ci is None:  # A is a vector
                c = A[ri]
                is_missing = self.missing[ri]
            else:  # A is a matrix
                c = A[ri][ci]
                is_missing = self.missing[ri] or self.missing[ci]

            if c is None:
                r = ""
            else:
                r = ", ".join(self._cell_label_html(printer, *r) for r in c)

            if is_missing:
                args.append('style="background-color: #%s;"' % self.missing_color)

        return "<td %s>%s</td>" % (" ".join(args), r)


class UnimolecularTable(_RxnTable):
    """Table of unimolecular reactions in a ReactionSystem

    Parameters
    ----------
    rsys : ReactionSystem
    sinks_sources_disjoint : tuple, None or True
        Colors sinks & sources. When ``True`` :meth:`sinks_sources_disjoint` is called.
    html_cell_label : Reaction formatting callback
        The function takes an integer, a Reaction instance and a dict of Substances as
        parameters and return a string.

    Returns
    -------
    string: html representation
    list: reactions not considered
    """

    _rsys_meth = "_unimolecular_reactions"

    def _html(self, printer, **kwargs):
        if "substances" not in kwargs:
            kwargs["substances"] = self.substances
        ss = printer._get("substances", **kwargs)
        rows = "\n".join(
            "<tr><td>%s</td>%s</tr>"
            % (printer._print(s), self._cell_html(printer, self.idx_rxn_pairs, rowi))
            for rowi, s in enumerate(ss.values())
        )
        return "<table>%s</table>" % rows


class BimolecularTable(_RxnTable):
    """Table of bimolecular reactions

    Parameters
    ----------
    idx_rxn_pairs : iterable of (int, Reaction) pairs
    substances : dict
        Mapping substance key to Substance instance.
    sinks_sources_disjoint : tuple, None or True
        Colors sinks & sources. When ``True`` :meth:`sinks_sources_disjoint` is called.

    Returns
    -------
    string: html representation
    list: reactions not considered
    """

    _rsys_meth = "_bimolecular_reactions"

    def _html(self, printer, **kwargs):
        if "substances" not in kwargs:
            kwargs["substances"] = self.substances
        ss = printer._get("substances", **kwargs)
        header = "<th></th>" + "".join(
            "<th>%s</th>" % printer._print(s) for s in ss.values()
        )
        rows = [
            "<tr><td>%s</td>%s</tr>"
            % (
                printer._print(s),
                "".join(
                    self._cell_html(printer, self.idx_rxn_pairs, rowi, ci)
                    for ci in range(len(ss))
                ),
            )
            for rowi, s in enumerate(ss.values())
        ]
        return "<table>%s</table>" % "\n".join([header, "\n".join(rows)])



# -*- coding: utf-8 -*-
"""
Utility functions related to stoichiometry.
"""


import numpy as np
from chempy.units import unit_of, to_unitless


def get_coeff_mtx(substances, stoichs):
    """
    Create a net stoichiometry matrix from reactions
    described by pairs of dictionaries.

    Parameters
    ----------
    substances : sequence of keys in stoichs dict pairs
    stoichs : sequence of pairs of dicts
        Pairs of reactant and product dicts mapping substance keys
        to stoichiometric coefficients (integers).

    Returns
    -------
    2 dimensional array of shape (len(substances), len(stoichs))

    """
    A = np.zeros((len(substances), len(stoichs)), dtype=int)
    for ri, sb in enumerate(substances):
        for ci, (reac, prod) in enumerate(stoichs):
            A[ri, ci] = prod.get(sb, 0) - reac.get(sb, 0)
    return A


def decompose_yields(yields, rxns, atol=1e-10):
    """Decomposes yields into mass-action reactions

    This function offers a way to express a reaction with non-integer
    stoichiometric coefficients as a linear combination of production reactions
    with integer coefficients.

    Ak = y

    A is (n_species x n_reactions) matrix, k is "rate coefficient", y is yields


    Parameters
    ----------
    yields : OrderedDict
        Specie names as keys and yields as values.
    rxns : iterable :class:`Reaction` instances
        Dict keys must match those of ``yields`` each pair
        of dictionaries gives stoichiometry
        (1st is reactant, 2nd is products).
    atol : float
        Absolute tolerance for residuals.


    Examples
    --------
    >>> from chempy import Reaction
    >>> h2a = Reaction({'H2O': 1}, {'H2': 1, 'O': 1})
    >>> h2b = Reaction({'H2O': 1}, {'H2': 1, 'H2O2': 1}, inact_reac={'H2O': 1})
    >>> decompose_yields({'H2': 3, 'O': 2, 'H2O2': 1}, [h2a, h2b])
    array([2., 1.])

    Raises
    ------
    ValueError
        When atol is exceeded
    numpy.LinAlgError
        When numpy.linalg.lstsq fails to converge

    Returns
    -------
    1-dimensional array of effective rate coefficients.

    """
    from chempy import ReactionSystem

    # Sanity check:
    rxn_keys = set.union(*(rxn.keys() for rxn in rxns))
    for key in yields.keys():
        if key not in rxn_keys:
            raise ValueError("Substance key: %s not in reactions" % key)
    rsys = ReactionSystem(rxns, rxn_keys)
    A = rsys.net_stoichs(yields.keys())
    b = list(yields.values())
    unit = unit_of(b[0])
    x, residuals, rank, s = np.linalg.lstsq(
        np.asarray(A.T, dtype=np.float64),
        to_unitless(b, unit),
        rcond=2e-16 * max(A.shape),
    )
    if len(residuals) > 0:
        if np.any(residuals > atol):
            raise ValueError("atol not satisfied")
    return x * unit



# -*- coding: utf-8 -*-

from .tables import (
    tables,
    symmetry_func_dict,
    column_coeffs,
    mulliken
)

"""
Contains chemical group theory functions for calculating symmetry adapted
linear combinations (SALCs) or group orbitals using either the projection
operator method or using the symmetry functions in the character tables.
"""

from math import cos, sin, radians, isclose
import numpy as np
from functools import wraps
import sympy
sympy.init_printing(pretty_print=False)


def return_dict(func):
    """
    Return results as a dictionary.

    Return a list or array as a dictionary with Mulliken symbols as
    the keys.

    Parameters
    ----------
    func : function returning SALCs
        Function that returns SALCs as a list by default.

    Returns
    -------
    Dictionary.

    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        if kwargs.get('to_dict'):
            keys = mulliken[args[1].lower()]
            values = func(*args, **kwargs)
            return dict(zip(keys, values))
        else:
            return func(*args, **kwargs)
    return wrapper


# PROJECTION OPERATOR METHOD

def _expand_irreducible(irred, group):
    """
    Return expanded irreducible.

    Expands irreducible from the common condensed form to the full form. This
    entails repeating characters for every equivalent operation. For example,
    C3v returns the rotation twice and the reflection thrice.

    Parameters
    ----------
    irred : tuple
        Condensed irreducible representation from character table.
    group : str
        Point group Schoenflies notation (e.g., 'C2v').  This is
        case-insensitive.

    Returns
    -------
    Expanded irreducible representation as a list.

    Example
    ------
    >>> _expand_irreducible((2, -1, 0), 'c3v')
    [2, -1, -1, 0, 0, 0]

    """
    expanded_irred = []
    for i in range(len(irred)):
        expanded_irred.extend([irred[i]] * column_coeffs[group.lower()][i])

    return expanded_irred


def _normalize_salcs_expr(salcs):
    """
    Normalize SALC composed of sympy expressions.

    Normalizes SALC by dividing each SALC through by the largest value in
    that SALC.

    Parameters
    ----------
    salcs : List or nested list of sympy expressions.
        Nested list of SALCs.

    Returns
    -------
    np.array

    """
    normalized_values = []
    for salc in salcs:
        if isinstance(salc, list):
            normalized_values.append(_normalize_salcs_expr(salc))
        elif salc == 0:
            normalized_values.append(salc)
        else:
            normalized_values.append(salc.as_poly().monic().as_expr())

    return normalized_values


@return_dict
def calc_salcs_projection(projection, group, to_dict=False):
    """
    Return SALCs using projection operator method.

    Given the projections of orbitals as a result of a point group symmetry
    operations, returns the SALCs. This is a two-step process.
    1. Provide all ligands or outer atoms with SymPy variable names.
    2. Track an orbital to see how it transforms after each symmetry operation
    3. Provide a list of the results from step 2.

    Note: The projection operator method only turns one SALC for E and T
    point groups.

    Parameters
    ----------
    projection : List, tuple, or array of SymPy symbols
        Results of projection operations for symmetry operations of group.
    group : str
        Point group Schoenflies notation (e.g., 'C2v').  This is
        case-insensitive.

    Returns
    -------
    List or nested list of strings of the SALCs for each irreducible
    representation. Returns 0 for irreducibles with no SALC. If to_dict=True,
    returns a dictionary.

    Example
    -------
    >>> import sympy
    >>> a, b, c = sympy.symbols('a b c')
    >>> calc_salcs_projection([a, b, c, a, b, c], 'c3v')
    [a + b + c, 0, a - b/2 - c/2]
    >>> calc_salcs_projection([a, b, c, a, b, c], 'c3v', to_dict=True)
    {'A1': a + b + c, 'A2': 0, 'E': a - b/2 - c/2}

    """
    salcs = []

    for irred in tables[group.lower()]:
        product = np.array(_expand_irreducible(irred, group.lower()) *
                           np.array(projection))
        salcs.append(np.sum(product))

    return _normalize_salcs_expr(salcs)


# USING SYMMETRY FUNCTIONS

def _angles_to_vectors(ligand_angles):
    """
    Calculate xyz vectors from angles around central atom.

    Given angles for outer ligands/atoms with respect to x-axis and z-axis,
    this function returns a list of [x, y, z] unit vectors.

    Parameters
    ----------
    ligand_angles : nested list or tuple of numbers in (theta, phi) order
        Angle of each outer atom/ligand in degrees with respect
        azimuth (theta) and polar (phi) coordinates. Theta is the angle from
        the positive x-axis on the xy-plane and phi is the angle from the
        positive z-axis.

    Returns
    -------
    Nested list containing xyz vectors.

    Example
    -------
    >>> _angles_to_vectors([[0, 90], [90, 90], [180, 90], [-90, 90]])
    [[1, 0, 0], [0, 1, 0], [-1, 0, 0], [0, -1, 0]]

    References
    ----------
    [1] https://en.wikipedia.org/wiki/Spherical_coordinate_system
    """
    all_vectors = []
    for orbital in ligand_angles:
        azimuth, polar = radians(orbital[0]), radians(orbital[1])
        x = sin(polar) * cos(azimuth)
        y = sin(polar) * sin(azimuth)
        z = cos(polar)

        # convert to int if close
        vector = []
        for val in (x, y, z):
            if isclose(val, int(val), abs_tol=1e-3):
                vector.append(int(val))
            else:
                vector.append(round(val, 3))

        all_vectors.append(vector)

    return all_vectors


def _eval_sym_func(coords, funcs):
    """
    Evaluate symmetry functions for an irreducible representation.

    Evaluates symmetry functions using the supplied series of xyz coordinates.
    If all values evaluate as zero, the irreducible has no SALC, and 0 is
    returned.

    Parameters
    ----------
    coords : List, tuple, or array containing values in threes
        xyz coordinates of ligand unit vectors.
    funcs : str
        The symmetry function supplied as a string or tuple of strings
        (e.g., 'x**2-y**2' or ('z**2', 'x**2+y**2')).

    Returns
    -------
    List or 0.

    """
    salcs = []

    for func in funcs:
        ligand_contribs = []
        for unit_vector in coords:
            x, y, z = unit_vector[0], unit_vector[1], unit_vector[2]
            ligand_contrib = eval(func, {'x': x, 'y': y, 'z': z})
            ligand_contribs.append(round(ligand_contrib, 2))

        if np.any(ligand_contribs):
            salcs.append(ligand_contribs)

    if not salcs:
        return 0
    elif len(salcs) == 1:
        return salcs[0]
    else:
        return salcs


def _normalize_salcs(salcs):
    """
    Normalize SALC.

    Normalizes SALC by dividing each SALC through by the largest value in
    that SALC.

    Parameters
    ----------
    salcs : List or nested list
        Nested list of SALCs.

    Returns
    -------
    np.array

    """
    normalized_values = []
    for value in salcs:
        if isinstance(value, list):
            normalized_values.append(_normalize_salcs(value))
        elif value == 0:
            normalized_values.append(value)
        else:
            coeff = round(value / max(salcs, key=abs), 2)
            if coeff % 1 < 0.01:
                coeff = sympy.Integer(coeff)
            normalized_values.append(coeff)

    return normalized_values


def _weights_to_symbols(weights, symbols):
    """
    Convert ligand weights to symbolic representations.

    Convert ligand or outer atom weights (e.g., [2, -0.5, -0.5]) to a symbolic
    representations (e.g., [2*a, -0.5*b, -0.5*c]).

    Parameters
    ----------
    weights : list
        List or nested list containing weights from each ligand or outer atom.
    symbols : list of Sympy symbols
        List of SymPy symbols the user provided to represent the ligands or
        outer atoms.

    Returns
    -------
    List containing symbolic weights of each ligand or outer atom.

    Examples
    --------
    >>> import sympy
    >>> a, b, c = sympy.symbols('a b c')
    >>> _weights_to_symbols([[1, 2, 1], [0, 1, -1], [0, 0, 0]], [a, b, c])
    [a + 2*b + c, b - c, 0]

    """
    symbolic_wt = []
    for weight in weights:
        if weight == 0:
            symbolic_wt.append(0)
        else:
            try:
                sym = np.array(weight).dot(np.array(symbols))
                if isinstance(sym, np.ndarray):
                    symbolic_wt.append(sym.tolist())
                else:
                    symbolic_wt.append(sym)
            except ValueError:
                symbolic_wt.append(_weights_to_symbols(weight, symbols))

    return symbolic_wt


@return_dict
def calc_salcs_func(ligands, group, symbols, mode='vector', to_dict=False):
    """
    Return SALCs using symmetry functions in character table.

    Returns SALCs (symmetry adapted linear combinations) using the character
    functions for a given point group. This requires the user to give the
    position of each ligand atomic orbital with respect to the x-axis, y-axis,
    and z-axis (mode='vector') or the spherical coordinate angles in degrees
    in pairs (azimuthal, polar) where the azimuthal is the angle from the
    positive x-axis on the xy-plane and polar is the angle from the positive
    z-axis (mode='angle').

    Parameters
    ----------
    ligand : list or nested list
        Nested list of ligand positions as xyz coordinates (mode='vector')
        or angles (mode='angle').
    group : str
        Point group Schoenflies notation (e.g., 'C2v').  This is
        case-insensitive.
    symbols : SymPy symbols
        SymPy symbols representing outer ligands or atoms.
    mode : 'vector' or 'angle'
        Whether the position of ligands or outer atoms are provided in xyz
        coordinates ('vector') or [theta, phi] angles ('angle').

    Returns
    -------
    List of sympy symbols indicating the weight and sign of each atomic
    orbital contribution to the SALC. There may be redundant SALCs returned
    due to multiple symmetry functions with an irreducible representation
    returning the same SALC.

    Examples
    --------
    >>> import sympy
    >>> a, b, c, d = sympy.symbols('a b c d')
    >>> coords = [[1, 0, 0], [0, 1, 0], [-1, 0, 0], [0, -1, 0]]
    >>> calc_salcs_func(coords, 'd4h', [a, b, c, d], mode='vector')
    [a + b + c + d, 0, a - b + c - d, 0, 0, 0, 0, 0, 0, [a - c, b - d]]
    >>> coords = [[0, -90], [120, -90], [240, -90]]
    >>> calc_salcs_func(coords, 'd3h', [a, b, c], mode='angle')
    [a + b + c, 0, [a - 0.5*b - 0.5*c, b - c, a - 0.5*b - 0.5*c, b - c], \
0, 0, 0]

    References
    ----------
    [1] Kim, S. K. Group Theoretical Methods and Applications to Molecules
    and Crystals; Cambridge University Press: Cambridge, 1999, 155-157.
    """
    if mode == 'angle':
        ligand_vectors = _angles_to_vectors(ligands)
    elif mode == 'vector':
        ligand_vectors = ligands
    else:
        raise Exception("Invalid mode input: must be 'angle' or 'vector'")

    salcs = []
    for sym_func in symmetry_func_dict[group]:
        if sym_func == 0:
            salcs.append(0)
        elif sym_func == 1:
            salcs.append(1)
        else:
            salcs.append(_eval_sym_func(ligand_vectors, sym_func))

    return _weights_to_symbols(_normalize_salcs(salcs), symbols)



# -*- coding: utf-8 -*-

from .numbers import (
    number_to_scientific_html,
    number_to_scientific_latex,
    number_to_scientific_unicode,
)
from .table import as_per_substance_html_table
from .js import javascript
from .string import str_
from .pretty import unicode_
from .web import html, css
from .tex import latex



# -*- coding: utf-8 -*-


import os
import subprocess
import shutil
import tempfile

from chempy import Reaction, ReactionSystem, Substance
from ..graph import rsys2dot, rsys2graph
from ..testing import requires, skipif

try:
    dot_missing = subprocess.run(["dot", "-?"]).returncode != 0
except OSError:
    dot_missing = True


def _get_rsys():
    r1 = Reaction({"A": 2}, {"B": 1}, param=3.0)
    A = Substance("A", latex_name="\\boldsymbol{A}")
    B = Substance("B", latex_name="\\boldsymbol{B}")
    rsys = ReactionSystem([r1], [A, B])
    return rsys


@requires("numpy")
@skipif(dot_missing, reason="graphviz not installed? (dot command missing)")
def test_rsys2dot():
    rsys = _get_rsys()
    assert list(map(str.strip, rsys2dot(rsys))) == [
        'digraph "None" {',
        '"A" [fontcolor=maroon label="A"];',
        '"B" [fontcolor=darkgreen label="B"];',
        "{",
        'node [label="r1",shape=diamond]',
        "r1",
        "}",
        '"A" -> "r1" [color=maroon,fontcolor=maroon,label="2"];',
        '"r1" -> "B" [color=darkgreen,fontcolor=darkgreen,label=""];',
        "}",
    ]


@requires("numpy")
@skipif(dot_missing, reason="graphviz not installed? (dot command missing)")
def test_rsys2graph():
    rsys = _get_rsys()
    tempdir = tempfile.mkdtemp()
    try:
        rsys2graph(rsys, os.path.join(tempdir, "out.png"))
        rsys2graph(rsys, os.path.join(tempdir, "out.ps"))
        try:
            subprocess.call(["dot2tex", "-v"])
        except Exception:
            pass
        else:
            rsys2graph(rsys, os.path.join(tempdir, "out.tex"))
    finally:
        shutil.rmtree(tempdir)



from itertools import chain


class Printer(object):
    _str = str  # set to unicdoe int UnicodePrinter for Python 2
    _default_settings = dict(
        with_param=True,
        with_name=True,
        fallback_print_fn=str,
        Reaction_param_separator="; ",
        Reaction_coeff_space=" ",
        Reaction_around_arrow=(" ", " "),
        magnitude_fmt=lambda x: "%.3g" % x,
    )
    _default_setting_factories = dict(
        substances=dict,
        colors=dict,  # substance key -> (bg-color, border-color), 6 char hex colors
    )
    _default_setting_attrs = dict(
        Reaction_coeff_fmt="_str",
        Reaction_formula_fmt="_str",
        unit_fmt="_str",
    )
    printmethod_attr = (
        None  # e.g. '_html' or '_unicode', allows object local printing logic
    )

    def __init__(self, settings=None):
        self._settings = dict(self._default_settings, **(settings or {}))
        for k, v in self._default_setting_factories.items():
            if k not in self._settings:
                self._settings[k] = v()
        for k, v in self._default_setting_attrs.items():
            if k not in self._settings:
                self._settings[k] = getattr(self, v)
        for k in self._settings:
            if k not in chain(
                self._default_settings,
                self._default_setting_factories,
                self._default_setting_attrs,
            ):
                raise ValueError(
                    "Unknown setting: %s (missing in default_settings)" % k
                )

    def _get(self, key, **kwargs):
        return kwargs.get(key, self._settings[key])

    def _print(self, obj, **kwargs):
        for cls in type(obj).__mro__:
            print_meth = "_print_" + cls.__name__
            if hasattr(self, print_meth):
                return getattr(self, print_meth)(obj, **kwargs)
            for PrintCls in self.__class__.__mro__:
                _attr = getattr(PrintCls, "printmethod_attr", None)
                if _attr and hasattr(obj, _attr):
                    return getattr(obj, _attr)(self, **kwargs)
        fn = self._get("fallback_print_fn", **kwargs)
        if fn:
            return fn(obj)
        else:
            raise ValueError("Don't know how to print obj of type: %s" % type(obj))

    def doprint(self, obj):
        return self._print(obj)



"""
Interactive kinetics app with sliders.
Start by running:
    $ bokeh serve interactive.py
Add --show argument or navigate to:
    http://localhost:5006/interactive
"""
from collections import defaultdict
import sys

from chempy import Reaction, ReactionSystem
from chempy.kinetics.rates import MassAction
from chempy.util.bkh import integration_with_sliders


def get_rsys(kf, kb):
    rf = MassAction([kf], unique_keys=["kf"])
    rb = MassAction([kb], unique_keys=["kb"])
    fw = Reaction({"Fe+3", "SCN-"}, {"FeSCN+2"}, rf)
    bw = Reaction({"FeSCN+2"}, {"Fe+3", "SCN-"}, rb)
    return ReactionSystem([fw, bw], "Fe+3 SCN- FeSCN+2".split())


if __name__.startswith("bk_"):
    from bokeh.io import curdoc

    curdoc().add_root(
        integration_with_sliders(
            get_rsys(3, 0.3),
            tend=3,
            c0=defaultdict(float, {"Fe+3": 0.9, "SCN-": 0.7, "FeSCN+2": 0.1}),
            parameters={"kf": 3, "kb": 0.3},
        )
    )
elif __name__ == "__main__":
    import warnings

    warnings.warn("Run using 'bokeh serve %s'" % __file__)
    sys.exit(1)



# -*- coding: utf-8 -*-

from ..regression import least_squares, irls
from ..testing import requires


@requires("numpy")
def test_irls():
    import numpy as np

    x = np.linspace(0, 100)
    y = np.exp(-x / 47)
    b, c, info = irls(x, np.log(y))
    assert abs(b[1] + 1 / 47) < 1e-5
    assert np.all(c < 1e-4)
    assert info["success"] is True
    assert info["niter"] < 3


@requires("numpy")
def test_least_squares():
    import numpy as np

    x, y, w = [0, 1, 2], [0, 1, 2], [1, 1, 1]
    beta, vcv, r2 = least_squares(x, y)
    assert np.allclose(beta, [0, 1])
    assert np.allclose(vcv, 0)
    assert np.allclose(r2, 1)

    wbeta, wvcv, wr2 = least_squares(x, y, w)
    assert np.allclose(wbeta, [0, 1])
    assert np.allclose(wvcv, 0)
    assert np.allclose(wr2, 1)



# -*- coding: utf-8 -*-

from collections import OrderedDict
import types
from ..pyutil import defaultkeydict, defaultnamedtuple, multi_indexed_cases


def test_defaultnamedtuple():
    Point2 = defaultnamedtuple("Point2", "x y", [10])
    p = Point2(3)
    assert p.x == 3 and p.y == 10

    Point3 = defaultnamedtuple("Point3", "x y z", [10, 20])
    p = Point3(3)
    assert p.x == 3 and p.y == 10 and p.z == 20

    p = Point3(3, z=30)
    assert p.x == 3 and p.y == 10 and p.z == 30

    p = Point3(3, 4, 5)
    assert p.x == 3 and p.y == 4 and p.z == 5

    _default_y = Point2.__new__.__defaults__[-1]

    class MySubclass(Point2):
        def __new__(cls, x2, y2=_default_y):
            return super(MySubclass, cls).__new__(cls, x2 ** 0.5, y2 ** 0.5)

    p2 = MySubclass(9, 4)
    assert isinstance(p2, tuple)
    assert isinstance(p2, Point2)
    assert isinstance(p2, MySubclass)
    assert not isinstance(p, MySubclass)
    assert p2.x == 3
    assert p2.y == 2

    p3 = MySubclass(9)
    assert p3.x == 3
    assert p3.y == 10 ** 0.5


def test_defaultkeydict():
    d = defaultkeydict(lambda k: k * 2)
    assert d["as"] == "asas"


def test_multi_indexed_cases():
    for mi, d in multi_indexed_cases([(7, "abc"), (8, "ef")]):
        assert type(d) is not dict
        assert isinstance(d, OrderedDict)
        assert isinstance(mi, tuple)

    result = multi_indexed_cases(
        [(97, ["0.5"]), (98, ["0.25"])],
        dict_=dict,
        apply_return=None,
        apply_values=float,
        apply_keys=chr,
        named_index=True,
    )
    assert isinstance(result, types.GeneratorType)
    ((mi, c),) = result
    assert mi == (0, 0)
    assert isinstance(mi, tuple)
    assert mi.a == 0
    assert mi.b == 0
    assert mi._asdict() == {"a": 0, "b": 0}
    assert type(c) is dict
    assert c == {"a": 0.5, "b": 0.25}
